# -*- coding: utf-8 -*-

"""Docstring Parsing Tools.

Docstring inheritance-style implementations.

To implement your own inheritance file, simply write a function that fits the template

.. code-block:: python

    def your_style(prnt_doc, child_doc):
        ''' Merges parent and child docstrings

            Parameters
            ----------
            prnt_cls_doc: Optional[str]
            child_doc: Optional[str]

            Returns
            ------
            Optional[str]
                The merged docstring that will be utilized.'''
        return final_docstring

and log this using `custom_inherit.add_style(your_style)`.
To permanently save your function

1. define your function within custom_inherit/_style_store.py
2. log it in custom_inherit.style_store.__all__.

Your style will then be available as 'your_style'
(i.e. whatever you named the function).

Routine Listings
----------------
merge_numpy_docs
merge_rest_docs
merge_numpy_napoleon_docs
merge_google_napoleon_docs

"""

__author__ = "Ryan Soklaski"
__license__ = "MIT"
__version__ = "2.2.2"
__email__ = "https://github.com/rsokl/custom_inherit"

__all__ = [
    "merge_numpy_docs",
    "merge_rest_docs",
    "merge_numpy_napoleon_docs",
    "merge_google_napoleon_docs",
]


##############################################################################
# IMPORTS

# PROJECT-SPECIFIC
from . import _style_store
from .napoleon_parse_tools import (
    merge_google_napoleon_docs,
    merge_numpy_napoleon_docs,
)
from .numpy_parse_tools import merge_numpy_docs
from .rest_parse_tools import merge_rest_docs

###############################################################################
# CODE
###############################################################################


def _check_style_function(style_func):
    out = style_func("", "")
    if not isinstance(out, str) and out is not None:
        raise TypeError
    return None


# /def


class _DocMergeStore:
    """A dictionary-like object that stores the docstring styles available

    Only callable objects with the signature::

        f(Optional[str], Optional[str]) -> Optional[str]

    can be stored. If f is a valid callable, then _DocMergeStore()[f] -> f.

    """

    def __init__(self, *args, **kwargs):
        self._store = dict()
        self.update(*args, **kwargs)

    # /def

    def __repr__(self):
        return repr(self._store)

    # /def

    def __str__(self):
        out_str = "The available stored styles are: "
        styles = "\n".join("\t- " + style for style in sorted(self.keys()))
        return "\n".join((out_str, styles))

    # /def

    def __setitem__(self, style_name, style_func):
        """Make available a new function for merging a 'parent' and 'child' docstring.

        Parameters
        ----------
        style_name : Any
            The identifier of the style being logged
        style_func: Callable[[Optional[str], Optional[str]], Optional[str]]
            The style function that merges two docstrings into a single docstring."""
        try:
            _check_style_function(style_func)
        except TypeError:
            raise TypeError(
                "The style store only stores callables of the form: "
                "\n\tstyle_func(Optional[str], Optional[str]) -> Optional[str]"
            )
        self._store[style_name] = style_func

    # /def

    def __getitem__(self, item):
        """Given a valid style-ID, retrieve a stored style.

        If a valid function (callable) is supplied, return it in place.

        Parameters
        ----------
        item : Union[Any, Callable[Optional[str], Optional[str]], Optional[str]]
            A valid style-ID or style-function.

        """
        try:
            return self._store[item]
        except KeyError:
            try:
                _check_style_function(item)
                return item
            except (TypeError, ValueError):
                raise TypeError(
                    "Either a valid style name or style-function must be specified"
                )

    # /def

    def keys(self):
        """ D.keys() -> a set-like object providing a view on D's keys"""
        return self._store.keys()

    # /def

    def pop(self, *args):
        """D.pop(k[,d]) -> v, remove specified key and return the corresponding value.

        If key is not found, d is returned if given,
        otherwise KeyError is raised.

        """
        if len(args) < 3:
            return self._store.pop(*args)
        else:
            raise TypeError(
                "pop expected at most 2 arguments, got {}".format(len(args))
            )

    # /def

    def update(self, *args, **kwargs):
        """D.update([E, ]**F) -> None.  Update D from dict/iterable E and F.

        If E is present and has a .keys() method, then does:  for k in E: D[k] = E[k]

        """
        if len(args) > 1:
            raise TypeError(
                "update expected at most 1 arguments, got %d" % len(args)
            )

        for key, value in dict(*args, **kwargs).items():
            self[key] = value

    # /def

    def values(self):
        """D.values() -> an object providing a view on D's values."""
        return self._store.values()

    # /def

    def items(self):
        """D.items() -> a set-like object providing a view on D's items"""
        return self._store.items()

    # /def


# /def

##############################################################################


store = _DocMergeStore(
    [(key, getattr(_style_store, key)) for key in _style_store.__all__]
)

##############################################################################


def add_style(style_name, style_func):
    """Make available a new function for merging a 'parent' and 'child' docstring.

    Parameters
    ----------
    style_name : Any
        The identifier of the style being logged
    style_func: Callable[[Optional[str], Optional[str]], Optional[str]]
        The style function that merges two docstrings into a single docstring.

    """
    store[style_name] = style_func


# /def


def remove_style(style):
    """Remove the specified style from the style store.

    Parameters
    ----------
    style: Any
        The inheritance-scheme style ID to be removed.

    """
    if style in store:
        store.pop(style)


# /def


##############################################################################
# END



# -*- coding: utf-8 -*-

"""Functions for interacting with the IPython environment.

Methods
-------
autoreload:
    set_autoreload: set the notebook's autoreload state for modules
    aimport: import a module with autoreload options
imports:
    import_from_file
    run_imports
    import_base, import_extended, import_astropy, import_matplotlib
    aimport from .autoreload
notebook:
    add_raw_code_toggle
plot:
    configure_matplotlib
printing:
    printmd / printMD
    printHTML
    printltx / printLaTeX

Returns
-------
IPython: modules and functions

    - get_ipython
    - core.interactiveshell

        - InteractiveShell
        - .debugger.set_trace

    - display.display, Latex, Markdown, HTML

utilipy: modules and functions

    - ipython

        - autoreload.set_autoreload, aimport
        - .imports.run_imports, import_from_file
        - .notebook.add_raw_code_toggle
        - .plot.configure_matplotlib
        - .printing.printmd, printMD, printltx, printLaTeX

    - utils.logging.LogPrint, LogFile

Notes
-----
will set the display setting to all output lines
and set the matplotlib backend to inline with retina resolution
`InteractiveShell.ast_node_interactivity = "all"`
`configure_matplotlib(backend='inline', figure_format='retina')`

References
----------
IPython [#]_

.. [#] Fernando Pérez, Brian E. Granger, IPython: A System for Interactive
    Scientific Computing, Computing in Science and Engineering, vol. 9,
    no. 3, pp. 21-29, May/June 2007, doi:10.1109/MCSE.2007.53.
    URL: https://ipython.org

"""

__author__ = "Nathaniel Starkman"


__all__ = [
    "ipython_help",
    "get_ipython",
    "InteractiveShell",
    "set_trace",
    "display",
    "Latex",
    "Markdown",
    "HTML",
    "set_autoreload",
    "aimport",
    "run_imports",
    "import_from_file",
    "add_raw_code_toggle",
    "printMD",
    "printLTX",
]


##############################################################################
# IMPORTS

# THIRD PARTY
from IPython import get_ipython
from IPython.core.debugger import set_trace
from IPython.core.interactiveshell import InteractiveShell
from IPython.display import HTML  # for printing Markdown & HTML
from IPython.display import Latex  # for printing LaTeX
from IPython.display import display  # display is a better print
from IPython.display import Markdown

# PROJECT-SPECIFIC
from .autoreload import aimport, set_autoreload
from .imports import import_from_file, run_imports
from .notebook import add_raw_code_toggle
from .plot import configure_matplotlib
from .printing import printLTX  # LaTeX printing
from .printing import printMD  # Markdown printing
from .setup_package import conf
from utilipy.utils import make_help_function

##############################################################################
# PARAMETERS

try:
    get_ipython()
    if get_ipython() is None:  # double checking
        raise NameError
except NameError:
    _HAS_IPY: bool = False

else:
    _HAS_IPY: bool = True

    # Running Imported Functions
    InteractiveShell.ast_node_interactivity = conf.ast_node_interactivity

    configure_matplotlib(
        backend=conf.ipython_matplotlib_backend,
        figure_format=conf.ipython_matplotlib_figure_format,
    )

# /try


##############################################################################
# HELP

ipython_help = make_help_function(
    "ipython", __doc__, look_for=None, doctitle="ipython module."
)


##############################################################################
# END



# -*- coding: utf-8 -*-

"""Test contents of :mod:`~utilipy.data_utils.fitting.lmfit_utils`."""


__all__ = [
    # functions
    "test_scipy_residual_to_lmfit_raises",
    "test_scipy_residual",
    "test_lmfit_residual",
]


##############################################################################
# IMPORTS

# THIRD PARTY
import numpy as np
import pytest

# PROJECT-SPECIFIC
from utilipy.data_utils.fitting.lmfit_utils import scipy_residual_to_lmfit

##############################################################################
# PARAMETERS


##############################################################################
# CODE
##############################################################################


def _model(x, amp, phaseshift, freq, decay):
    return amp * np.sin(x * freq + phaseshift) * np.exp(-(x ** 2) * decay)


# /def


@scipy_residual_to_lmfit.decorator(
    param_order=["amp", "phaseshift", "freq", "decay"]
)
def _residual(variables, x, data, eps_data):
    """Model a decaying sine wave and subtract data."""
    amp = variables[0]
    phaseshift = variables[1]
    freq = variables[2]
    decay = variables[3]

    model = _model(x, amp=amp, freq=freq, phaseshift=phaseshift, decay=decay)

    return (data - model) / eps_data


# /def

# -------------------------------------------------------------------


def test_scipy_residual_to_lmfit_raises():
    """Test Exceptions `~utilipy.data_utils.fitting.scipy_residual_to_lmfit`."""
    with pytest.raises(ValueError):
        scipy_residual_to_lmfit(param_order=None)


# /def

# -------------------------------------------------------------------


def test_scipy_residual():
    """"""
    x = np.arange(0, 10)
    variables = [1, 2, 3, 4]  # amp, phaseshift, freq, decay

    data = _model(x, *variables)
    eps_data = 1.0

    assert all(_residual(variables, x=x, data=data, eps_data=eps_data) == 0.0)


# /def

# -------------------------------------------------------------------


def test_lmfit_residual():
    """"""
    x = np.arange(0, 10)

    class _ParameterProxy:
        def __init__(self, name, value):
            self.name = name
            self.value = value

    class _ParametersProxy(dict):
        def __init__(self, *params):
            for p in params:
                self[p.name] = p

        def add_many(self, params_list):
            for (n, v) in params_list:
                self[n] = _ParameterProxy(n, v)

        def valuesdict(self):
            return {n: p.value for n, p in self.items()}

    params = _ParametersProxy()
    params.add_many([("amp", 1), ("phaseshift", 2), ("freq", 3), ("decay", 4)])

    data = _model(x, **params.valuesdict())
    eps_data = 1.0

    assert all(
        _residual.lmfit(params, x=x, data=data, eps_data=eps_data) == 0.0
    )


# /def


##############################################################################
# END



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# ----------------------------------------------------------------------------
#
# TITLE   : LogFile
# PROJECT : utilipy
#
# ----------------------------------------------------------------------------

"""Functions for logging."""

__author__ = "Nathaniel Starkman"


##############################################################################
# IMPORTS

# BUILT-IN
import typing as T

# PROJECT-SPECIFIC
from ._LogPrint import LogPrint
from utilipy.utils.metaclasses import InheritDocstrings

##############################################################################
# LogFile


class LogFile(LogPrint, metaclass=InheritDocstrings):
    """Class for basic logger that can both print and record to a file.

    this class uses `open`, not a more extensive logger, like `logging`

    The arguments `mode` to `opener` are all for ``open``
    their descriptions are in
    `<https://docs.python.org/3/library/functions.html#open>`_

    Parameters
    ----------
    filename : str, optional
        the file name / path at which to save this log
        If no filename, makes a ``LogPrint`` instead
    verbose : int, optional
        the verbosity level to use in ``.report``
    mode : str, optional
        recommend either 'w' or 'a'
    sec_div : str, optional
        the section divider used in `newsection`
    header : None or str, optional
        the header for the file
        None goes to filename
    show_header : bool, optional
        whether to print the header

    Notes
    -----
    mode options
        * 'r' open for reading
        * 'w' open for writing, truncating the file first
        * 'x' open for exclusive creation, failing if the file already exists
        * 'a' open for writing, appending to the end of the file if it exists
        * 'b' binary mode
        * 't' text mode
        * NOT ALLOWED '+' open a disk file for updating (reading and writing)

    """

    def __new__(
        cls,
        filename: T.Optional[str] = None,
        verbose: int = 0,
        sec_div: str = "-",
        header: T.Union[str, None, bool] = None,
        show_header: bool = True,
        # for open
        mode: str = "w",
        buffering: int = -1,
        encoding: None = None,
        errors: None = None,
        newline: None = None,
        closefd: bool = True,
        opener: None = None,
    ) -> None:
        """LogFile.

        if no filename, makes a LogPrint instead

        """
        if mode == "+":
            raise ValueError("+ not allowed")

        if filename is None:
            return LogPrint(
                verbose=verbose,
                sec_div=sec_div,
                header=header,
                show_header=show_header,
            )
        # else:
        self = super().__new__(cls)
        return self

    # /def

    def __init__(
        self,
        filename: T.Optional[str] = None,
        verbose: int = 0,
        sec_div: str = "-",
        header: T.Union[str, None, bool] = None,
        show_header: bool = True,
        # for open
        mode: str = "w",
        buffering: int = -1,
        encoding: None = None,
        errors: None = None,
        newline: None = None,
        closefd: bool = True,
        opener: None = None,
    ) -> None:
        """LogFile.

        set the filename and make the file

        """
        # instantiate without writing
        # section divider and file header
        super().__init__(verbose=verbose, sec_div=sec_div, header=False)

        # keeping input arguments
        self.filename = filename

        # the Logger
        self.file = open(
            filename,
            mode=mode,
            buffering=buffering,
            encoding=encoding,
            errors=errors,
            newline=newline,
            closefd=closefd,
            opener=opener,
        )

        if mode == "r":
            return

        # making file header
        if header is False:  # no header
            self.write(f"{''} Log:", endsection="=", print=False)
        else:  # there is a header
            if header is None:  # header should be filename
                header = filename + " "
            elif header is True:  # blank header
                header = ""
            elif header[-1] != " ":  # making sure ends in space
                header += " "

            self.write(f"{header}Log:", endsection="=", print=show_header)

        return

    # /def

    def __getattr__(self, name: str) -> T.Any:
        """Redirect non-defined attributes to self.file."""
        return getattr(self.file, name)

    # /def

    # ------------------------------------------------------------------------

    @classmethod
    def open(
        cls,
        filename: str,
        verbose: int = 0,
        sec_div: str = "-",
        header: T.Optional[str] = None,
        show_header: bool = True,
        # for open
        mode: str = "w",
        buffering: int = -1,
        encoding: None = None,
        errors: None = None,
        newline: None = None,
        closefd: bool = True,
        opener: None = None,
    ) -> T.Any:
        """Make basic logger that can both print and record to a file.

        this class uses `open`, not a more extensive logger, like `logging`

        The arguments `mode` to `opener` are all for ``open``
        their descriptions are in
        `<https://docs.python.org/3/library/functions.html#open>`_

        Parameters
        ----------
        filename: str
            the file name / path at which to save this log
        mode: str
            recommend either 'w' or 'a'
        sec_div: str
            the section divider used in `newsection`
        header : None or str
            the header for the file
            None -> filename

        Notes
        -----
        mode options
            * 'r' open for reading
            * 'w' open for writing, truncating the file first
            * 'x' open for exclusive creation, failing if file already exists
            * 'a' open for writing, appending to the end of file if it exists
            * 'b' binary mode
            * 't' text mode
            * '+' open a disk file for updating (reading and writing)

        """
        return cls(
            filename,
            verbose=verbose,
            mode=mode,
            sec_div=sec_div,
            header=header,
            show_header=show_header,
            buffering=buffering,
            encoding=encoding,
            errors=errors,
            newline=newline,
            closefd=closefd,
            opener=opener,
        )

    # /def

    @classmethod
    def open_to_write(
        cls,
        filename: str,
        verbose: int = 0,
        sec_div: str = "-",
        header: T.Optional[str] = None,
        show_header: bool = True,
        # for open
        mode: int = "w",
        buffering: int = -1,
        encoding: None = None,
        errors: None = None,
        newline: None = None,
        closefd: bool = True,
        opener: None = None,
    ) -> T.Any:
        """Make basic logger that can both print and record to a file.

        this class uses `open`, not a more extensive logger, like `logging`

        The arguments `mode` to `opener` are all for ``open``
        their descriptions are in
        `<https://docs.python.org/3/library/functions.html#open>`_

        Parameters
        ----------
        filename: str
            the file name / path at which to save this log
        mode: str
            recommend either 'w' or 'a'
            cannot be 'r'
        sec_div: str
            the section divider used in `newsection`
        header : None or str, optional
            the header for the file
            None -> filename

        Notes
        -----
        mode options
            * 'w' open for writing, truncating the file first
            * 'x' open for exclusive creation, failing if file already exists
            * 'a' open for writing, appending to the end of file if it exists
            * 'b' binary mode
            * 't' text mode
            * '+' open a disk file for updating (reading and writing)

        """
        if mode == "r":
            raise ValueError("mode must be set to write")

        return cls(
            filename,
            verbose=verbose,
            mode=mode,
            sec_div=sec_div,
            header=header,
            show_header=show_header,
            buffering=buffering,
            encoding=encoding,
            errors=errors,
            newline=newline,
            closefd=closefd,
            opener=opener,
        )

    # /def

    @classmethod
    def open_to_read(
        cls,
        filename: str,
        # for open
        buffering: int = -1,
        encoding: None = None,
        errors: None = None,
        newline: None = None,
        closefd: bool = True,
        opener: None = None,
    ) -> T.Any:
        """Make basic logger that can both print and record to a file.

        this class uses `open`, not a more extensive logger, like `logging`

        The arguments `buffering` to `opener` are all for ``open``
        their descriptions are in
        `<https://docs.python.org/3/library/functions.html#open>`_

        Parameters
        ----------
        filename: str
            the file name / path at which to save this log

        """
        return cls(
            filename,
            mode="r",
            buffering=buffering,
            encoding=encoding,
            errors=errors,
            newline=newline,
            closefd=closefd,
            opener=opener,
        )

    # /def

    # ------------------------------------------------------------------------

    def _write(
        self, *string: str, start: str = "", sep: str = " ", end: str = "\n"
    ):
        r"""Write helper method.

        this is used by all write methods
        implemented so it can be overriden easily
        Note: end='' does nothing. Write automatically does '\n'

        """
        if len(string) == 0:  # checking there is a string
            raise ValueError("needs a value")

        self.file.write(start)  # start

        # write to file
        if len(string) == 1:
            self.file.write(str(string[0]) + end)

        else:
            for s in string[:-1]:  # all strings with sep
                self.file.write(str(s) + sep)
            self.file.write(str(string[-1]) + end)  # last string

        self.file.flush()  # force write to file
        return

    # /def

    def write(
        self,
        *text: str,
        start: str = "",
        sep: str = " ",
        end: str = "\n",
        startsection: bool = False,
        endsection: bool = False,
        print: bool = True,
    ) -> T.Any:
        r"""Write string to stream and print it to output.

        Parameters
        ----------
        text: str
            the text to write & print
        start: str
            start to print
        sep: str
            the separater for print
        end: str
            the end for print
        startsection: bool
            whether to start a new section before writing
        endsection:  bool

        """
        return super().write(
            *text,
            start=start,
            sep=sep,
            end=end,
            startsection=startsection,
            endsection=endsection,
            print=print,
        )

    # /def

    def record(
        self,
        *text: str,
        start: str = "",
        end: str = "\n",
        startsection: bool = False,
        endsection: bool = False,
    ) -> T.Any:
        """Write, but doesn't print as well as write to file."""
        return super().record(
            *text,
            start=start,
            end=end,
            startsection=startsection,
            endsection=endsection,
        )

    # /def

    # ------------------------------------------------------------------------

    def report(
        self,
        *msgs: str,
        verbose: T.Optional[int] = None,
        print: bool = True,
        write: bool = True,
        start_at: int = 1,
        **kw,
    ) -> T.Any:
        """a report function whose message is determined by the `verbose`.

        Parameters
        ----------
        msgs : str(s)
            the verbosity-ordered messages
            blank messages can be <None>, not only ''
        verbose : int, optional
            which message to record
            None (default) uses self.verbose (default = 0, unless specified)
        print : bool
            whether to print, or just record
        write : bool
            whether to write to logger file
        start_at : int
            what level of verbosity is the first `msgs`
            ex: verbort('test', start_at=3) means 'test' is at verbose=3
        kw: kwargs for self.write or self.print

        """
        return super().report(
            *msgs,
            verbose=verbose,
            print=print,
            write=write,
            start_at=start_at,
            **kw,
        )

    # /def

    # ------------------------------------------------------------------------

    def close(self):
        """Close the file."""
        self.newsection(title="closing file", div="=")
        self.file.close()
        return

    # /def


# /class

##############################################################################
# END



# -*- coding: utf-8 -*-

"""Collections."""

__author__ = "Nathaniel Starkman"


__all__ = [
    # -------
    "WithDocstring",
    "WithMeta",
    "WithReference",
    # -------
    "MetaDataBase",
    "ReferenceBase",
    # -------
    "ObjDict",
]


##############################################################################
# IMPORTS

# BUILT-IN
import typing as T
from collections import OrderedDict

# THIRD PARTY
import wrapt
from astropy.utils.decorators import format_doc
from astropy.utils.metadata import MetaData

# PROJECT-SPECIFIC
from .pickle import dump as _dump  # nosec
from .pickle import load as _load  # nosec

##############################################################################
# PARAMETERS

odict_values = type(OrderedDict().values())
odict_items = type(OrderedDict().items())


##############################################################################
# CODE
##############################################################################

##########################################################################
# Wrapper Classes


class WithDocstring(wrapt.ObjectProxy):
    """Wrap object to add docstring.

    This is a :class:`~wrapt.ObjectProxy`.

    - Thin: All operations are carried to the wrapt value,
        except for ``__doc__``, which accesses the docstring attribute.
    - Light: operations do not carry the docstring

        >>> x = WithDocstring(2, doc="description")
        >>> y = x + 2

    Attributes
    ----------
    __doc__ : str
        description.

    """

    __doc__ = None

    def __init__(self, wrapped: T.Any, doc: str = None):
        """Object wrapt to include metadata.

        Parameters
        ----------
        wrapped : T.Any
            `WithMeta` wraps anything, just adding metadata.
        kw : Dict[T.Any, T.Any]
            the key, value pairs for the metadata

        """
        super().__init__(wrapped)
        self.__doc__ = doc

        return

    # /def


# /class


# ---------------------------------------------------------------------------


class WithMeta(wrapt.ObjectProxy):
    """Objects wrapt to include metadata.

    This is an :class:`~wrapt.ObjectProxy` for thin and light wrappers.

    - Thin: All operations are carried to the wrapt value,
            except for ``self.meta``, which accesses the meta attribute.
    - Light: operations do not carry the metadata

        >>> x = WithMeta(2)
        >>> y = x + 2
        >>> y.meta  # doctest: +SKIP
        AttributeError: 'int' object has no attribute 'meta'

    Attributes
    ----------
    meta : `~astropy.utils.metadata.MetaData`
        metadata OrderedDict managed by the Astropy metadata module.

        .. todo::

            implement with actual MetaData descriptor, not OrderedDict


    Notes
    -----
    see :mod:`~astropy.utils.metadata` for usage of the metadata module.

    """

    # meta = MetaData(copy=False)
    _meta: T.Optional[T.Dict] = None

    def __init__(self, wrapped: T.Any, **kw):
        """Object wrapt to include metadata.

        Parameters
        ----------
        wrapped : T.Any
            `WithMeta` wraps anything, just adding metadata.
        kw : Dict[T.Any, T.Any]
            the key, value pairs for the metadata

        """
        super().__init__(wrapped)
        # self.meta = MetaData(copy=False)
        self._meta = OrderedDict()

        for k, v in kw.items():
            self._meta[k] = v

    # /def

    @property
    def meta(self):
        """Metadata."""
        return self._meta

    # /def


# /class


# ---------------------------------------------------------------------------


class WithReference(WithMeta, WithDocstring):
    """Objects wrapt to include reference(s) and other metadata.

    Attributes
    ----------
    meta : `~astropy.utils.metadata.MetaData`
        metadata OrderedDict managed by the Astropy metadata module.
        Always has a "reference" item; None if not set.

    Notes
    -----
    see :mod:`~astropy.utils.metadata` for usage of the metadata module.

    """

    def __init__(
        self,
        wrapped: T.Any,
        *,
        doc: str = None,
        reference: T.Optional[T.Any] = None,
        **kw,
    ):
        """Object wrapt to include a reference, and any other metadata.

        Parameters
        ----------
        wrapped : Any
            `WithMeta` wraps anything, just adding metadata.
        reference: Any, optional
            The attribution for the object, such as a paper link.
        kw : Dict[Any, Any], any
            the key, value pairs for the metadata

        """
        # instantiate, with metadata
        # need to do this first to make the wrapt object
        super().__init__(wrapped, reference=reference, **kw)
        # then manually do WithDocstring
        self.__doc__ = doc

        return

    # /def

    @property
    def __reference__(self):
        """References Meta Information."""
        return self.meta["reference"]

    # /def


# /class


##########################################################################
# Base Classes


class MetaDataBase:
    r"""Superclass for object with MetaData.

    Child classes accept the class argument ``meta_copy``.

    Attributes
    ----------
    meta : `~astropy.utils.metadata.MetaData`
        Meta-data descriptor with ``copy`` set by `_meta_copy`.
    \_meta_copy : bool
        Whether `meta` has ``copy=True``.
        This is set with the class argument ``meta_copy``

        >>> class Child(MetaDataBase, meta_copy=False):
        ...     pass

    Examples
    --------
    Making a child class

        >>> class Child(MetaDataBase, meta_copy=False):
        ...     pass

    """

    _meta = OrderedDict()

    def __init_subclass__(cls, meta_copy: bool = True):
        """Control subclass creation.

        Parameters
        ----------
        meta_copy : bool, optional
            Whether the `~astropy.utils.metadata.MetaData` instance uses
            ``copy=True``.
            The value is stored in ``self._meta_copy``

        """
        cls.meta = MetaData(copy=meta_copy)
        cls._meta_copy = meta_copy
        super().__init_subclass__()

    # /def

    def __new__(cls, *args, **kwargs):
        if cls == MetaDataBase:
            raise TypeError(f"Must subclass {cls.__name__} to use.")
        return super().__new__(cls)

    # /def

    def __init__(self, **metadata):
        """Initialize Metadata.

        Parameters
        ----------
        **metadata : Any
            added to ``meta``

        """
        super().__init__()
        self.meta.update(metadata)

    # /def


# /class


# -------------------------------------------------------------------


class ReferenceBase(MetaDataBase):
    """Superclass for objects with references.

    Accepts the class argument ``meta_copy``.

    Attributes
    ----------
    meta : `~astropy.utils.metadata.MetaData`
        Meta-data descriptor with ``copy=False``.
        Always has a key "reference"
    __reference__  : Any
        A property attribute to access the key-value "reference" in `meta`.

    """

    def __new__(cls, *args, **kwargs):
        if cls == ReferenceBase:
            raise TypeError(f"Must subclass {cls.__name__} to use.")
        return super().__new__(cls, *args, **kwargs)

    # /def

    def __init__(self, *, reference: T.Union[None, str] = None, **metadata):
        """Initialize Reference Metadata.

        Parameters
        ----------
        reference : str, optional
            reference, added to metadata.
            accessible by ``__reference__``
        **metadata
            To add to metadata.

        """
        super().__init__(reference=reference, **metadata)

    # /def

    @property
    def __reference__(self):
        """Get reference from metadata, if exists."""
        return self.meta.get("reference", None)

    # /def


# /class


##########################################################################
# ObjDict


class ObjDict(OrderedDict):
    """Dictionary-like object intended to store information.

    instantiated with a name (str)
    supports __getattr__ as a redirect to __getitem__.

    Parameters
    ----------
    name: str
        the name of the object
    **kw
        items for ObjDict

    Examples
    --------
    >>> obj = ObjDict('NAME', a=1, b=2)
    >>> print(obj.name, obj.a, obj['b'])
    NAME 1 2

    """

    def __init__(self, name: str = "", **kw: T.Any):
        """Initialize ObjDict."""
        super().__init__()

        object.__setattr__(self, "name", name)

        # TODO better fix for inspect._is_wrapper
        object.__setattr__(self, "__wrapped__", None)

        for key, value in kw.items():
            self[key] = value

        return

    # /def

    # ----------------------------------
    # item get / set

    def __getitem__(
        self, keys: T.Union[str, T.Sequence[str]], as_generator: bool = False
    ) -> T.Any:
        """Override __getitem__.

        Parameters
        ----------
        keys: str, list of str
            the keys into ObjDict
            if str: just the value from key-value pair
            if list: list of values
        as_generator: bool  (default False)
            whether to return as a generator
            only if keys is a list

        Returns
        -------
        value(s): anything
            if str, just the value from key-value pair
            if list, list of values
            if as_generator, generator for list

        Examples
        --------
        obj = ObjDict('NAME', a=1, b=2)
        print(obj['a'])
        >> 1, [NAME, 2]
        print(obj['name', 'b'])
        >> [NAME, 2]

        """
        # single key
        if isinstance(keys, str):
            return super().__getitem__(keys)
        # multiple keys
        if as_generator:  # return generator
            return (OrderedDict.__getitem__(self, k) for k in keys)
        return [OrderedDict.__getitem__(self, k) for k in keys]

    # /def

    @format_doc(__getitem__.__doc__)
    def getitem(
        self, keys: T.Union[str, T.Sequence[str]], as_generator: bool = False
    ) -> T.Any:
        """Docstring from __getitem__."""
        return self.__getitem__(keys, as_generator=as_generator)

    # /def

    # ----------------------------------
    # attribute get / set
    # redirects to item get / set

    def __getattr__(self, key: T.Any) -> T.Any:
        """Getattr -> getitem."""
        return self[key]

    # /def

    def __setattr__(self, key: T.Any, value: T.Any) -> None:
        """Setattr -> setitem."""
        self[key] = value

    # /def

    # ----------------------------------
    # Printing

    def __repr__(self) -> str:
        """__repr__."""
        if self.name == "":
            return super().__repr__()
        else:
            return self.name + super().__repr__().replace("ObjDict", "")

    # /def

    # ----------------------------------

    def values(self, *keys: T.Any) -> T.Union[odict_values, tuple]:
        """Values.

        Parameters
        ----------
        *keys: keys for values subset, default is all keys

        Returns
        -------
        values: list
            list of values for `*keys`

        """
        if keys:  # keys provided
            allkeys = self.keys()
            return tuple([self[k] for k in allkeys if k in keys])
        return super().values()

    # /def

    def items(self, *keys: T.Any) -> T.Union[odict_items, dict]:
        """Items.

        Parameters
        ----------
        *keys: keys for items subset, default is all keys

        Returns
        -------
        items: dictionary
            items for `*keys`

        """
        if not keys:
            return super().items()
        return self.fromkeys(keys).items()

    # /def

    def fromkeys(self, keys: T.Any = [], name: str = None) -> T.Any:
        """Subset.

        Parameters
        ----------
        *keys: keys for subset dictionary, default is all keys

        Returns
        -------
        subset: list
            items for `*keys`

        """
        if not keys:
            return self
        else:
            allkeys = self.keys()

            if name is None:
                name = self.name + " subset"

            return ObjDict(
                name=name, **{k: self[k] for k in allkeys if k in keys}
            )

    # /def

    def keyslist(self) -> list:
        """Keys in list form.

        Parameters
        ----------
        *keys: keys for subset dictionary, default is all keys

        Returns
        -------
        subset: list
            items for `*keys`

        """
        return list(self.keys())

    # /def

    # ----------------------------------
    # Serialize (I/O)

    def __reduce__(self) -> T.Tuple[T.Callable, T.Tuple[str], odict_items]:
        """Reduction method for serialization.

        structured as:
        1. the class
        2. (ObjDict name, )
        3. items

        """
        return (self.__class__, (self.name,), OrderedDict(self.items()))

    # /def

    def __setstate__(self, state: dict) -> None:
        """Set-state method for loading from pickle.

        sets by key, value pairs

        """
        for key, value in state.items():
            self[key] = value

    # /def

    def dump(
        self,
        fname: str,
        protocol: T.Optional[int] = None,
        *,
        fopt: str = "b",
        fix_imports: bool = True,
    ):
        """Dump to pickle file.

        uses :func:`~utilipy.utils.pickle.dump`

        """
        _dump(
            self, fname, protocol=protocol, fopt=fopt, fix_imports=fix_imports
        )

    # /def

    @format_doc(dump.__doc__)
    def save(
        self,
        fname: str,
        protocol: T.Optional[int] = None,
        *,
        fopt: str = "b",
        fix_imports: bool = True,
    ):
        """.Dump alias."""
        self.dump(fname, protocol=protocol, fopt=fopt, fix_imports=fix_imports)

    # /def

    @staticmethod
    def load(
        fname: str,
        *,
        fopt: str = "b",
        fix_imports: bool = True,
        encoding: str = "ASCII",
        errors: str = "strict",
    ):
        """Load from pickle file.

        uses :func:`~utilipy.utils.pickle.load`

        """
        self = _load(
            fname,
            fopt=fopt,
            fix_imports=fix_imports,
            encoding=encoding,
            errors=errors,
        )
        return self

    # /def

    # ----------------------------------

    def print(self):
        """Print."""
        print(self.__repr__())

    # /def


# /class


##############################################################################
# END



# -*- coding: utf-8 -*-

# ----------------------------------------------------------------------------
#
# TITLE   : galpy_imports
# PROJECT : utilipy
#
# ----------------------------------------------------------------------------

# Docstring and Metadata
"""Base set of imports for :mod:`~galpy`.

Routine Listings
----------------
Galpy : imports

    - potential, .MWPotential2014
    - galpy.orbit.Orbit
    - galpy.util: bovy_conversion, bovy_coords

References
----------
Galpy reference [#]_.

.. [#] galpy: A Python Library for Galactic Dynamics, Jo Bovy (2015),
    Astrophys. J. Supp., 216, 29 (arXiv/1412.3451)

"""

__author__ = "Nathaniel Starkman"


__all__ = [
    "galpy_imports_help",
]


##############################################################################
# HELPER FUNCTIONS

# PROJECT-SPECIFIC
from utilipy.imports import conf
from utilipy.utils import make_help_function

##############################################################################
# IMPORTS

try:

    import galpy

except ImportError:

    import warnings

    warnings.warn("Cannot import galpy")

else:

    # potential
    from galpy import potential
    from galpy.potential import MWPotential2014

    # orbit
    from galpy.orbit import Orbit

    # util
    from galpy.util import bovy_conversion, bovy_coords

    from galpy.util import galpyWarning

    __all__ += [
        # modules
        "galpy",
        # functions
        "potential",
        "MWPotential2014",
        "Orbit",
        "bovy_conversion",
        "bovy_coords",
        "galpyWarning",
    ]

# /if

##############################################################################
# Printing Information

galpy_imports_help = make_help_function(
    "galpy", __doc__, look_for="Routine Listings"
)


if conf.verbose_imports:
    galpy_imports_help()


##############################################################################
# END



# -*- coding: utf-8 -*-

"""Tests for :mod:`~utilipy.utils.pickle`."""


__all__ = [
    "test_config",
    "Test_dump",
    "test_dump_many",
    "test_load",
]


##############################################################################
# IMPORTS

# BUILT-IN
import tempfile

# THIRD PARTY
import numpy as np
import pytest

# PROJECT-SPECIFIC
from utilipy.tests.helper import BaseClassDependentTests
from utilipy.utils import pickle

##############################################################################
# PARAMETERS

objs = {
    int: 1,
    float: 0.2,
    list: [1, 2],
    tuple: (3, 4),
    dict: {"a": 1, "b": 2},
    np.ndarray: np.array([1, 2, 3, 4, 5]),
}

##############################################################################
# TESTS
##############################################################################


def test_config():
    """Test `~utilipy.utils.pickle.conf`.

    .. todo::

        A unified method of testing configs

    """
    assert pickle.conf.use_dill is False

    with pickle.conf.set_temp("use_dill", True):
        assert pickle.conf.use_dill is True


# /def


# --------------------------------------------------------------------------


class Test_dump(BaseClassDependentTests, klass=pickle.dump):
    """Test `~utilipy.utils.pickle.dump`."""

    @pytest.mark.parametrize("obj", objs)
    @staticmethod
    def test_can_dump_objs(obj):
        """Test can dump a few different objects. Unpickle later."""
        # temporary file
        with tempfile.NamedTemporaryFile() as fname:

            pickle.dump(obj, fname=fname)

    # /def

    @pytest.mark.skipif(not pickle.HAS_DILL, reason="`dill` not installed.")
    @pytest.mark.parametrize("obj", objs)
    @staticmethod
    def test_use_dill(obj):
        """Test can dump a few different objects with dill."""
        # temporary file
        with tempfile.NamedTemporaryFile() as fname:

            pickle.dump(obj, fname=fname, use_dill=True)

    # /def

    @pytest.mark.skipif(pickle.HAS_DILL, reason="`dill` installed.")
    @staticmethod
    def test_fail_use_dill():
        """Test failed use of dill.

        Skip if has dill, since want it to fail.
        """
        # expect error
        with pytest.raises(ValueError):

            # make temporary file
            with tempfile.NamedTemporaryFile() as fname:

                pickle.dump([1, 2, 3], fname=fname, use_dill=True)

    # /def


# /class

# --------------------------------------------------------------------------


@pytest.mark.skip(reason="TODO")
def test_dump_many():
    """Test :func:`~utilipy.utils.pickle.dump_many`."""
    assert False


# /def


# --------------------------------------------------------------------------


@pytest.mark.skip(reason="TODO")
def test_load():
    """Test :func:`~utilipy.utils.pickle.load`."""
    assert False


# /def


##############################################################################
# END



# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""Setup for :mod:`~utilipy.ipython`."""


###############################################################################
# IMPORTS

from __future__ import absolute_import

# BUILT-IN
import typing as T

# THIRD PARTY
from astropy import config as _config

###############################################################################
# CODE
###############################################################################


class Conf(_config.ConfigNamespace):
    """Configuration parameters for `utilipy.ipython`."""

    # AST

    ast_node_interactivity = _config.ConfigItem(
        "all",
        description="AST node interactivity.",
        cfgtype="string(default='all')",
    )

    # Matplotlib

    ipython_matplotlib_backend = _config.ConfigItem(
        "inline",
        description="Configure IPython matplotlib backend.",
        cfgtype="string(default='inline')",
    )

    ipython_matplotlib_figure_format = _config.ConfigItem(
        "retina",
        description="Configure IPython matplotlib figure format.",
        cfgtype="string(default='retina')",
    )


conf = Conf()
# /class


###############################################################################
# INFO


__all__: T.List[str] = [
    "conf",  # configuration instance
]


###############################################################################
# END



# -*- coding: utf-8 -*-

"""Tests for `~utilipy.tests.helper.BaseClassDependentTests`."""

__all__ = [
    "Test_BaseClassDependentTests",
]


##############################################################################
# IMPORTS

# THIRD PARTY
import pytest

# PROJECT-SPECIFIC
from utilipy.tests import helper

##############################################################################
# CODE
##############################################################################


class Test_BaseClassDependentTests:
    """Test `~utilipy.tests.helper.BaseClassDependentTests`."""

    @classmethod
    def setup_class(cls):
        """Setup fixtures for testing."""
        cls.klass = helper.BaseClassDependentTests

    # /def

    def test_fail_make_subclass(self):
        """Test make subclass without passing klass."""
        with pytest.raises(TypeError) as e:
            # make class
            class SubClass(helper.BaseClassDependentTests):
                pass

        assert "klass" in str(e.value)

    # /def

    def test_make_subclass(self):
        """Test make subclass with klass."""
        # make class
        class SubClass(helper.BaseClassDependentTests, klass=int):
            pass

        # tests
        assert SubClass.klass is int

        subclass = SubClass()
        assert subclass.klass == int

    # /def

    def test_fail_make_subsubclass(self):
        """Test make subclass without passing klass."""
        # make class
        class SubClass(helper.BaseClassDependentTests, klass=int):
            pass

        with pytest.raises(TypeError) as e:

            class SubSubClass(SubClass):
                pass

        assert "klass" in str(e.value)

    # /def

    def test_make_subsubclass(self):
        """Test make sub-subclass with klass."""
        # make class
        class SubClass(helper.BaseClassDependentTests, klass=int):
            pass

        # make subclass
        class SubSubClass(SubClass, klass=int):
            pass

        # tests
        assert SubSubClass.klass is int

        subsubclass = SubSubClass()
        assert subsubclass.klass == int

    # /def


# /class


# -------------------------------------------------------------------


##############################################################################
# END



# -*- coding: utf-8 -*-

"""Test contents of :mod:`~utilipy.utils.misc`."""

__all__ = [
    "test_temporary_namespace",
    "test_make_help_function_raises",
]


##############################################################################
# IMPORTS

# BUILT-IN
import types

# THIRD PARTY
import pytest

# PROJECT-SPECIFIC
from utilipy.utils import misc

##############################################################################
# PARAMETERS

_doc_ = __doc__

_doc = """Docstring

Returns
-------
Nothing.

"""


##############################################################################
# CODE
##############################################################################


def test_temporary_namespace():  # TODO FIX
    """Test :func:`~utilipy.utils.misc.temporary_namespace`."""
    x = 1
    y = 2

    with misc.temporary_namespace(locals(), keep=["a"]):
        locals()["a"] = 3  # TODO FIX
        locals()["b"] = 4

    assert x, y == (1, 2)
    assert locals()["a"] == 3

    with pytest.raises(AssertionError):
        assert "b" in locals()


# /def


# -------------------------------------------------------------------


def test_make_help_function_raises():
    """Test :func:`~utilipy.utils.misc.make_help_function` Exceptions."""
    with pytest.raises(TypeError):
        misc.make_help_function("test", look_for=Exception)


# /def


def test_make_help_function_find_module():
    """Test :func:`~utilipy.utils.misc.make_help_function` Exceptions."""
    help_function = misc.make_help_function("test", module=None)

    assert help_function.__name__ == "test_help"
    assert help_function.__module__ == "utilipy.utils.tests.test_misc"
    assert help_function.__doc__ == "Help for test.\n\n" + _doc_


# /def


def test_make_help_function_pass_module():
    """Test :func:`~utilipy.utils.misc.make_help_function` Exceptions."""
    module = types.ModuleType("null_module", doc=_doc)

    help_function = misc.make_help_function("test", module=module)

    assert help_function.__name__ == "test_help"
    assert help_function.__module__ == "null_module"
    assert help_function.__doc__ == "Help for test.\n\n" + _doc


# /def


##############################################################################
# END



# -*- coding: utf-8 -*-

"""Test functions for AMUSE imports."""

__all__ = ["test_import_amuse"]


##############################################################################
# IMPORTS

# BUILT-IN
import warnings

# THIRD PARTY
import pytest

try:
    import amuse
except ImportError:
    HAS_AMUSE = False
else:
    if isinstance(amuse.__version__, str):  # check not empty namespace
        HAS_AMUSE = True
    else:
        HAS_AMUSE = False


##############################################################################


@pytest.mark.skipif(~HAS_AMUSE, reason="requires AMUSE")
def test_import_amuse():
    """Test AMUSE imports."""
    warnings.filterwarnings("ignore", category=DeprecationWarning)
    from utilipy.imports import amuse_imports as imports

    for obj in (
        "amuse",
        "lab",
        "units",
        "constants",
        "bridge",
    ):
        assert hasattr(imports, obj)


# /def


##############################################################################
# END



# -*- coding: utf-8 -*-

"""Tests for :mod:`~utilipy.tests.quantity_array.quantity_array`."""

__all__ = [
    "Test_eltwise_quantity_isclose",
    "Test_eltwise_quantity_allclose",
    "Test_eltwise_assert_quantity_isclose",
    "Test_eltwise_assert_quantity_allclose",
]


##############################################################################
# IMPORTS

# THIRD PARTY
import astropy.units as u
import numpy as np
import pytest

# PROJECT-SPECIFIC
from utilipy.tests.helper import quantity_array  # BaseClassDependentTests,

##############################################################################
# CODE
##############################################################################


class Test_eltwise_quantity_isclose:
    """Test :func:`~utilipy.tests.helper.eltwise_quantity_isclose`."""

    @property
    def func(self):
        """Tested function."""
        return quantity_array.eltwise_quantity_isclose

    @pytest.mark.parametrize(
        "a,b,rtol,atol,expected", [(1, 1 + 1e-16, 1e-15, None, True)]
    )
    def test_nounit(self, a, b, rtol, atol, expected):
        """Test when don't pass units."""
        assert self.func(a, b, rtol=rtol, atol=atol) == expected

    # /def

    @pytest.mark.parametrize(
        "a,b,rtol,atol,expected", [(1 * u.m, 1 * u.s, 1e-15, None, True)]
    )
    def test_wrongunit(self, a, b, rtol, atol, expected):
        """Test when pass wrong units."""
        with pytest.raises(u.UnitsError):
            assert self.func(a, b, rtol, atol)

    # /def

    @pytest.mark.parametrize(
        "a,b,rtol,atol,expected",
        [
            (1 * u.m, 1 * u.m, 1e-15, None, True),
            (0 * u.m, 0 * u.m, 1e-15, None, True),  # basic
            (1e-17 * u.m, 0 * u.m, 0, 1e-15 * u.m, True),  # close to 0
            (0 * u.m, 1e-17 * u.m, 0, 1e-15 * u.m, True),  # close to 0
            (0 * u.m, 1 * u.m, 1e-15, None, False),  # fails
            (0 * u.m, 1 * u.m, 2, None, True),  # adjust rtol to make work
        ],
    )
    def test_scalar(self, a, b, rtol, atol, expected):
        """Test when arguments are scalars.

        .. todo::

            Add hypothesis tests

        """
        assert self.func(a, b, rtol=rtol, atol=atol) == expected

    # /def

    @pytest.mark.parametrize(
        "a,b,rtol,atol,expected",
        [
            ([0, 0] * u.m, [0, 1] * u.m, 1e-15, None, [True, False]),  # basic
            ([1e-17, 0] * u.m, [0, 0] * u.m, 0, 1e-15 * u.m, [True, True]),
            ([1, 0] * u.m, [0, 1e-17] * u.m, 0, 1e-15 * u.m, [False, True]),
            ([0, 1] * u.m, [1, 1] * u.m, 1e-15, None, [False, True]),  # fails
            ([0, 1] * u.m, [1, 2] * u.m, 2, None, [True, True]),
            (  # units IN the list
                [0 * u.m, 1 * u.s],
                [0 * u.m, 1 * u.s],
                1e-15,
                None,
                [True, True],
            ),
            (
                [0 * u.m, 1 * u.s],
                [0 * u.m, 2 * u.s],
                1e-15,
                None,
                [True, False],
            ),
        ],
    )
    def test_vector(self, a, b, rtol, atol, expected):
        """Tests when arguments are vectors.

        To address [#148]

        .. todo::

            Add hypothesis tests

        """
        assert np.all(self.func(a, b, rtol=rtol, atol=atol) == expected)

    # /def


# /class

# -------------------------------------------------------------------


class Test_eltwise_quantity_allclose(Test_eltwise_quantity_isclose):
    """Test :func:`~utilipy.tests.helper.eltwise_quantity_allclose`."""

    @property
    def func(self):
        """Tested function."""
        return quantity_array.eltwise_quantity_allclose

    @pytest.mark.parametrize(
        "a,b,rtol,atol,expected",
        [
            ([0, 0] * u.m, [0, 0] * u.m, 1e-15, None, True),  # basic
            ([1e-17, 0] * u.m, [0, 0] * u.m, 0, 1e-15 * u.m, True),
            ([0, 0] * u.m, [1e-17, 0] * u.m, 0, 1e-15 * u.m, True),
            ([0, 1] * u.m, [1, 1] * u.m, 1e-15, None, False),  # fails
            ([0, 1] * u.m, [1, 2] * u.m, 2, None, True),
            # units IN the list
            ([0 * u.m, 1 * u.s], [0 * u.m, 1 * u.s], 1e-15, None, True),
        ],
    )
    def test_vector(self, a, b, rtol, atol, expected):
        """Tests when arguments are vectors.

        Unit tests IN the list are from [#148].

        .. todo::

            Add hypothesis tests

        """
        assert self.func(a, b, rtol=rtol, atol=atol) == expected

    # /def

    @pytest.mark.parametrize(
        "a,b,rtol,atol",
        [
            # units mismatch
            ([0 * u.m, 1 * u.s], [0 * u.m, 2 * u.m], 1e-15, None),
        ],
    )
    def test_vector_fails(self, a, b, rtol, atol):
        """Tests when arguments are vectors.

        Unit tests IN the list are from [#148].

        .. todo::

            Add hypothesis tests

        """
        with pytest.raises(u.UnitsError):
            self.func(a, b, rtol=rtol, atol=atol)

    # /def


# /class


# -------------------------------------------------------------------


class Test_eltwise_assert_quantity_isclose:
    """Test `~utilipy.tests.helper.eltwise_assert_quantity_isclose`.

    test no-units, wrong units, and scalar inputs.

    """

    @property
    def func(self):
        """Tested function."""
        return quantity_array.eltwise_assert_quantity_isclose

    @pytest.mark.parametrize(
        "a,b,rtol,atol",
        [
            (1, 1 + 1e-16, 1e-15, None),
        ],
    )
    def test_nounit(self, a, b, rtol, atol):
        """Test when don't pass units."""
        self.func(a, b, rtol=rtol, atol=atol)

    # /def

    @pytest.mark.parametrize(
        "a,b,rtol,atol",
        [(1 * u.m, 1 * u.s, 1e-15, None)],
    )
    def test_wrongunit(self, a, b, rtol, atol):
        """Test when pass wrong units."""
        with pytest.raises(u.UnitsError):
            self.func(a, b, rtol, atol)

    # /def

    @pytest.mark.parametrize(
        "a,b,rtol,atol",
        [
            (1 * u.m, 1 * u.m, 1e-15, None),
            (0 * u.m, 0 * u.m, 1e-15, None),  # basic
            (1e-17 * u.m, 0 * u.m, 0, 1e-15 * u.m),  # close to 0
            (0 * u.m, 1e-17 * u.m, 0, 1e-15 * u.m),  # close to 0
            (0 * u.m, 1 * u.m, 2, None),  # adjust rtol to make work
        ],
    )
    def test_scalar(self, a, b, rtol, atol):
        """Test when arguments are scalars.

        .. todo::

            Add hypothesis tests

        """
        self.func(a, b, rtol=rtol, atol=atol)

    # /def

    @pytest.mark.parametrize(
        "a,b,rtol,atol",
        [
            (0 * u.m, 1 * u.m, 1e-15, None),
        ],  # fails
    )
    def test_scalar_fails(self, a, b, rtol, atol):
        """Test when arguments are scalars.

        .. todo::

            Add hypothesis tests

        """
        with pytest.raises(AssertionError):
            self.func(a, b, rtol=rtol, atol=atol)

    # /def

    @pytest.mark.parametrize(
        "a,b,rtol,atol",
        [
            ([0, 0] * u.m, [0, 0] * u.m, 1e-15, None),
            ([1e-17, 0] * u.m, [0, 0] * u.m, 0, 1e-15 * u.m),
            ([0, 0] * u.m, [1e-17, 0] * u.m, 0, 1e-15 * u.m),
            ([0, 1] * u.m, [1, 2] * u.m, 2, None),
        ],
    )
    def test_vector(self, a, b, rtol, atol):
        """Tests when arguments are vectors.

        To address [#148]

        .. todo::

            Add hypothesis tests

        """
        self.func(a, b, rtol=rtol, atol=atol)

    # /def

    @pytest.mark.parametrize(
        "a,b,rtol,atol",
        [([0, 1] * u.m, [1, 1] * u.m, 1e-15, None)],
    )
    def test_vector_fails(self, a, b, rtol, atol):
        """Tests when arguments are vectors.

        .. todo::

            Add hypothesis tests

        """
        with pytest.raises(AssertionError):
            self.func(a, b, rtol=rtol, atol=atol)

    # /def


# /class


# -------------------------------------------------------------------


class Test_eltwise_assert_quantity_allclose(
    Test_eltwise_assert_quantity_isclose
):
    """Test `~utilipy.tests.helper.eltwise_assert_quantity_isclose`.

    Based on the Test_eltwise_assert_quantity_isclose tests.

    """

    @property
    def func(self):
        """Tested function."""
        return quantity_array.eltwise_assert_quantity_allclose

    @pytest.mark.parametrize(
        "a,b,rtol,atol",
        [
            ([0, 0] * u.m, [0, 0] * u.m, 1e-15, None),
            ([1e-17, 0] * u.m, [0, 0] * u.m, 0, 1e-15 * u.m),
            ([0, 0] * u.m, [1e-17, 0] * u.m, 0, 1e-15 * u.m),
            ([0, 1] * u.m, [1, 2] * u.m, 2, None),
            # units IN the list
            ([0 * u.m, 1 * u.s], [0 * u.m, 1 * u.s], 1e-15, None),
        ],
    )
    def test_vector(self, a, b, rtol, atol):
        """Tests when arguments are vectors.

        Unit tests IN the list are from [#148].

        .. todo::

            Add hypothesis tests

        """
        self.func(a, b, rtol=rtol, atol=atol)

    # /def

    @pytest.mark.parametrize(
        "a,b,rtol,atol",
        [
            ([0, 1] * u.m, [1, 1] * u.m, 1e-15, None),
            # units IN the list
            ([0 * u.m, 1 * u.s], [0 * u.m, 2 * u.m], 1e-15, None),
        ],
    )
    def test_vector_fails(self, a, b, rtol, atol):
        """Tests when arguments are vectors.

        Unit tests IN the list are from [#148].

        .. todo::

            Add hypothesis tests

        """
        # Raises AssertionError or UnitsError
        with pytest.raises((AssertionError, u.UnitsError)):
            self.func(a, b, rtol=rtol, atol=atol)

    # /def


# /class

# -------------------------------------------------------------------


##############################################################################
# END



# -*- coding: utf-8 -*-

"""Tests for module :mod:`~utilipy.decorators.code_dev`."""

__all__ = [
    "test_DevelopmentWarning",
    "test_BetaDevelopmentWarning",
    "test_indev_doc",
    "test_indev_func_message",
    "test_indev_decorator_func",
    "test_indev_decorator_class",
]


##############################################################################
# IMPORTS

# BUILT-IN
import warnings

# THIRD PARTY
import pytest

# PROJECT-SPECIFIC
from utilipy.decorators import code_dev as cdev

##############################################################################
# CODE
##############################################################################

#####################################################################
# Warnings


def test_DevelopmentWarning():
    """Test `~utilipy.decorators.code_dev.DevelopmentWarning`."""
    with pytest.warns(cdev.DevelopmentWarning):

        warnings.warn(cdev.DevelopmentWarning())

    # /with


# /def


# -------------------------------------------------------------------


def test_BetaDevelopmentWarning():
    """Test `~utilipy.decorators.code_dev.BetaDevelopmentWarning`."""
    with pytest.warns(cdev.BetaDevelopmentWarning):

        warnings.warn(cdev.BetaDevelopmentWarning())

    # /with


# /def


#####################################################################
# Indev


def test_indev_doc():
    """Test :func:`~utilipy.decorators.code_dev.indev_doc`."""
    old_doc = """Short Summary.

    Parameters
    ----------
    x, y : float

    Returns
    -------
    None

    """

    message = "test code dev"

    new_doc = cdev.indev_doc(old_doc=old_doc, message=message)

    expected_doc = (
        "Short Summary.\n\n"
        "    .. versionchanged:: indev\n\n        test code dev\n\n"
        "    Parameters\n    ----------\n    x, y : float\n\n"
        "    Returns\n    -------\n    None"
    )

    assert new_doc == expected_doc


# /def


# -------------------------------------------------------------------


def test_indev_func_message():
    """Test :func:`~utilipy.decorators.code_dev.indev` function message."""
    # test function
    def _func():
        """Short Summary."""

    # /dev

    indev_func = cdev.indev(
        message=_func,
        name="",
        alternative="",
        todo="",
        obj_type=None,
        warning_type=cdev.DevelopmentWarning,
    )

    assert indev_func.__wrapped__ == _func

    expected_doc = (
        "Short Summary.\n\n"
        "    .. versionchanged:: indev\n\n"
        "        The _func function is in development "
        "and may be added in a future version.\n\n"
    )

    assert indev_func.__doc__ == expected_doc


# /def


# -------------------------------------------------------------------


def test_indev_decorator_func():
    """Test :func:`~utilipy.decorators.code_dev.indev` function message."""
    # --------------

    @cdev.indev()
    def func():
        """Short Summary."""
        return "result"

    # /dev

    assert hasattr(func, "__wrapped__")
    assert func() == func.__wrapped__() == "result"

    assert func.__wrapped__.__doc__ == "Short Summary."
    assert func.__doc__ == (
        "Short Summary.\n\n"
        "    .. versionchanged:: indev\n\n"
        "        The func function is in development "
        "and may be added in a future version.\n\n"
    )

    with pytest.warns(cdev.DevelopmentWarning):
        warnings.warn(func())
    # /with

    # --------------

    @cdev.indev(message="indev message")
    def func():
        """Short Summary."""

    # /dev

    assert func.__doc__ == (
        "Short Summary.\n\n"
        "    .. versionchanged:: indev\n\n"
        "        indev message\n\n"
    )

    # --------------

    @cdev.indev(name="new_name")
    def func():
        """Short Summary."""

    # /dev

    assert func.__doc__ == (
        "Short Summary.\n\n"
        "    .. versionchanged:: indev\n\n"
        "        The new_name function is in development "
        "and may be added in a future version.\n\n"
    )

    # --------------

    @cdev.indev(alternative="alternative")
    def func():
        """Short Summary."""

    # /dev

    assert func.__doc__ == (
        "Short Summary.\n\n"
        "    .. versionchanged:: indev\n\n"
        "        The func function is in development "
        "and may be added in a future version.\n"
        "        Use alternative instead.\n\n"
    )

    # --------------

    @cdev.indev(todo="test")
    def func():
        """Short Summary."""

    # /dev

    assert func.__doc__ == (
        "Short Summary.\n\n"
        "    .. versionchanged:: indev\n\n"
        "        The func function is in development "
        "and may be added in a future version.\n"
        "        TODO: test\n\n"
    )

    # --------------

    @cdev.indev(warning_type=cdev.BetaDevelopmentWarning)
    def func():
        """Short Summary."""

    # /dev

    assert func.__doc__ == (
        "Short Summary.\n\n"
        "    .. versionchanged:: indev\n\n"
        "        The func function is in development "
        "and may be added in a future version.\n\n"
    )

    with pytest.warns(cdev.BetaDevelopmentWarning):
        warnings.warn(func())
    # /with

    # --------------

    @cdev.indev(obj_type="FUNCTION")
    def func():
        """Short Summary."""

    # /dev

    assert func.__doc__ == (
        "Short Summary.\n\n"
        "    .. versionchanged:: indev\n\n"
        "        The func FUNCTION is in development "
        "and may be added in a future version.\n\n"
    )

    # --------------

    @cdev.indev(warning_type=UserWarning)
    def func():
        """Short Summary."""

    # /dev

    with pytest.warns(UserWarning):
        warnings.warn(func())
    # /with


# /def


# -------------------------------------------------------------------


def test_indev_decorator_class():
    """Test :func:`~utilipy.decorators.code_dev.indev` class.

    Don't need to a lot of tests on the docstring, as these are covered
    by both :func:`~utilipy.decorators.code_dev.tests.test_indev_doc` and
    :func:`~utilipy.decorators.code_dev.tests.test_indev_decorator_func`.

    """
    # ----------------

    @cdev.indev()
    class Class(object):
        """docstring for Class."""

        def __init__(self):
            """Short Summary."""
            super(Class, self).__init__()

    # /class

    assert Class.__doc__ == (
        "docstring for Class.\n\n"
        "    .. versionchanged:: indev\n\n"
        "        The Class class is in development "
        "and may be added in a future version.\n\n"
    )

    assert hasattr(Class.__init__, "__wrapped__")

    assert Class.__init__.__wrapped__.__doc__ == "Short Summary."

    with pytest.warns(cdev.DevelopmentWarning):
        warnings.warn(Class())
    # /with


# /def

##############################################################################
# END



# -*- coding: utf-8 -*-

"""Function Input and Output Decorators."""

__all__ = [
    "store_function_input",
    "add_folder_backslash",
    "random_generator_from_seed",
    "dtypeDecoratorMaker",
]


##############################################################################
# IMPORTS

# BUILT-IN
import typing as T

# THIRD PARTY
import numpy as np

# PROJECT-SPECIFIC
from utilipy.utils import functools, inspect
from utilipy.utils.typing import EllipsisType

##############################################################################
# CODE
##############################################################################


def store_function_input(
    function: T.Callable = None,
    *,
    store_inputs: bool = True,
    _doc_style: str = "numpy",
    _doc_fmt: T.Dict[str, T.Any] = {},
):
    """Store Function Inputs.

    Store the function inputs as a BoundArguments.

    Parameters
    ----------
    function : T.Callable or None, optional
        the function to be decoratored
        if None, then returns decorator to apply.
    store_inputs : bool, optional
        whether to return all the inputs to the function as a dictionary

    Returns
    -------
    wrapper : T.Callable
        Wrapper for `function` that can store function inputs
        in a BoundArguments instance.
        Includes the original function in a method `.__wrapped__`

    Other Parameters
    ----------------
    _doc_style: str or formatter, optional
        default 'numpy'
        parameter to `utilipy.wraps`
    _doc_fmt: dict, optional
        default None
        parameter to `utilipy.wraps`

    """
    if function is None:  # allowing for optional arguments
        return functools.partial(
            store_function_input,
            store_inputs=store_inputs,
            _doc_style=_doc_style,
            _doc_fmt=_doc_fmt,
        )

    sig = inspect.fuller_signature(function)
    _doc_fmt["wrapped_function"] = function.__name__

    @functools.wraps(function, _doc_style=_doc_style, _doc_fmt=_doc_fmt)
    def wrapper(*args, store_inputs: bool = store_inputs, **kw):
        """Wrapper docstring.

        Parameters
        ----------
        store_inputs: bool
            whether to store function inputs in a `~inspect.BoundArguments`
            instance default {store_inputs}

        Returns
        -------
        inputs: `~inspect.BoundArguments`
            the inputs to ``{wrapped_function}``
            only returned if `store_inputs` is True
            other returned values are in now in a tuple

        """
        return_ = function(*args, **kw)

        if store_inputs:  # make and return BoundArguments
            inputs = sig.bind_partial_with_defaults(*args, **kw)
            return return_, inputs
        else:
            return return_

    # /def

    return wrapper


# /def


# -------------------------------------------------------------------


def add_folder_backslash(
    function=None,
    *,
    arguments: T.List[T.Union[str, int]] = [],
    _doc_style="numpy",
    _doc_fmt={},
):
    """Add backslashes to str arguments.

    For use in ensuring directory file-paths end in '/', when
    ``os.join`` just won't do.

    Parameters
    ----------
    function : T.Callable or None, optional
        the function to be decoratored
        if None, then returns decorator to apply.
    arguments : list of string or int, optional
        arguments to which to append '/', if not already present
        strings are names of arguments.
        Can also be int, which only applies to args.

    Returns
    -------
    wrapper : T.Callable
        wrapper for function
        does a few things
        includes the original function in a method ``.__wrapped__``

    Other Parameters
    ----------------
    _doc_style: str or formatter, optional
        default 'numpy'
        parameter to `~utilipy.wraps`
    _doc_fmt: dict, optional
        default None
        parameter to `~utilipy.wraps`

    Examples
    --------
    For modifying a single argument

    >>> @add_folder_backslash(arguments='path')
    ... def func(path):
    ...     return path
    >>> func("~/Documents")
    '~/Documents/'

    When several arguments need modification.

    >>> @add_folder_backslash(arguments=('path1', 'path2'))
    ... def func(path1, path2):
    ...     return (path1, path2)
    >>> func("~/Documents", "~Desktop")
    ('~/Documents/', '~Desktop/')

    """
    if isinstance(arguments, (str, int)):  # recast as tuple
        arguments = (arguments,)

    if function is None:  # allowing for optional arguments
        return functools.partial(
            add_folder_backslash,
            arguments=arguments,
            _doc_style=_doc_style,
            _doc_fmt=_doc_fmt,
        )

    sig = inspect.signature(function)

    @functools.wraps(function, _doc_style=_doc_style, _doc_fmt=_doc_fmt)
    def wrapper(*args, **kw):
        """Wrapper docstring.

        Parameters
        ----------
        store_inputs: bool
            whether to store function inputs in a BoundArguments instance
            default {store_inputs}

        """
        # bind args & kwargs to function
        ba = sig.bind_partial(*args, **kw)
        ba.apply_defaults()

        for name in arguments:  # iter through args
            # first check it's a string
            if not isinstance(ba.arguments[name], (str, bytes)):
                continue
            else:
                str_type = type(ba.arguments[name])  # get string type

            backslash = str_type("/")  # so we can work with any type

            if isinstance(name, int):  # only applies to args
                if not ba.args[name].endswith(backslash):
                    ba.args[name] += backslash
            elif isinstance(name, str):  # args or kwargs
                if not ba.arguments[name].endswith(backslash):
                    ba.arguments[name] += backslash
            else:
                raise TypeError("elements of `args` must be int or str")

        return function(*ba.args, **ba.kwargs)

    # /def

    return wrapper


# /def


#####################################################################


def random_generator_from_seed(
    function: T.Callable = None,
    seed_names: T.Union[str, T.Sequence[str]] = ("random", "random_seed"),
    generator: T.Callable = np.random.RandomState,
    raise_if_not_int: bool = False,
):
    """Function decorator to convert random seed to random number generator.

    Parameters
    ----------
    function : types.FunctionType or None (optional)
        the function to be decoratored
        if None, then returns decorator to apply.
    seed_names : list (optional)
        possible parameter names for the random seed
    generator : ClassType (optional)
        ex :class:`numpy.random.default_rng`, :class:`numpy.random.RandomState`

    raise_if_not_int : bool (optional, keyword-only)
        raise TypeError if seed argument is not an int.

    Returns
    -------
    wrapper : types.FunctionType
        wrapper for function
        converts random seeds to random number generators before calling.
        includes the original function in a method `.__wrapped__`

    Raises
    ------
    TypeError
        If `raise_if_not_int` is True and seed argument is not an int.

    """
    if isinstance(seed_names, str):  # correct a bare string to list
        seed_names = (seed_names,)

    if function is None:  # allowing for optional arguments
        return functools.partial(
            random_generator_from_seed,
            seed_names=seed_names,
            generator=generator,
        )

    sig = inspect.signature(function)
    pnames = tuple(sig.parameters.keys())

    @functools.wraps(
        function,
        _doc_fmt={"seed_names": seed_names, "random_generator": generator},
    )
    def wrapper(*args, **kw):
        """Wrapper docstring, added to Function.

        Notes
        -----
        T.Any argument in {seed_names} will be interpreted as a random seed,
        if it is an integer, and will be converted to a random number generator
        of type {random_generator}.

        """
        ba = sig.bind_partial(*args, **kw)
        ba.apply_defaults()

        # go through possible parameter names for the random seed
        # if it is a parameter and the value is an int, change to RandomState
        for name in seed_names:  # iterate through possible
            if name in pnames:  # see if present
                if isinstance(ba.arguments[name], int):  # seed -> generator
                    ba.arguments[name] = generator(ba.arguments[name])
                elif raise_if_not_int:
                    raise TypeError(f"{name} must be <int>")
                else:  # do not replace
                    pass
        # /for

        return function(*ba.args, **ba.kwargs)

    # /def

    return wrapper


# /def


#####################################################################
# DataType Decorators


class dtypeDecorator:
    """Ensure arguments are type *dtype*.

    Parameters
    ----------
    func : function, optional
        function to decorate
    inargs : list
        [(index, dtype), ...]

    outargs : list
        [(index, dtype), ...]

    these arguments, except func, should be specified by key word
    if inargs is forgotten and func is not a function, then func is
    assumed to be inargs.

    """

    def __new__(
        cls,
        func: T.Optional[T.Callable] = None,
        in_dtype: T.Any = None,
        out_dtype: T.Any = None,
    ):
        """New dtypeDecorator."""
        self = super().__new__(cls)  # making instance of self

        # correcting if forgot to specify in_dtype and no function
        # in this case, *in_dtype* is stored in *func*
        # need to do func->None, inarags<-func, and out_dtype<-in_dtype
        if not isinstance(func, T.Callable):
            # moving arguments 'forward'
            out_dtype = in_dtype
            in_dtype = func
            func = None

        # allowing for wrapping with calling the class
        if func is not None:
            self.__init__(in_dtype, out_dtype)
            return self(func)
        # else:
        return self

    # /def

    def __init__(
        self, in_dtype: T.Any = None, out_dtype: T.Any = None
    ) -> None:
        """Initialize dtypeDecorator."""
        super().__init__()

        # in_dtype
        # TODO check in_dtype is list of lists
        self._in_dtype = in_dtype

        # out_dtype
        # TODO check out_dtype is list of lists
        self._out_dtype = out_dtype

        return

    # /def

    def __call__(self, wrapped_func: T.Callable) -> T.Callable:
        """Make Decorator.

        Parameters
        ----------
        wrapped_func: Callable
            function to be wrapped

        """
        # make wrapper
        @functools.wraps(wrapped_func)
        def wrapper(*args: T.Any, **kw: T.Any) -> T.Any:
            # making arguments self._dtype
            if self._in_dtype is None:  # no conversion needed
                return_ = wrapped_func(*args, **kw)
            elif len(args) == 0:
                return_ = wrapped_func(**kw)
            elif len(args) == 1:
                # TODO better
                if len(self._in_dtype) != 1 or self._in_dtype[0][0] != 0:
                    raise IndexError("too many indices")
                arg = self._in_dtype[0][1](args[0])
                return_ = wrapped_func(arg, **kw)
            else:
                args = list(args)  # allowing modifications
                for i, dtype in self._in_dtype:
                    args[i] = dtype(args[i])  # converting to desired dtype`
                return_ = wrapped_func(*args, **kw)

            # POST
            if self._out_dtype is None:  # no conversion needed
                pass
            else:
                if not np.isscalar(return_):
                    return_ = list(return_)  # allowing modifications
                    for i, dtype in self._out_dtype:
                        # converting to desired dtype
                        return_[i] = dtype(return_[i])
                else:
                    if len(self._out_dtype) != 1:  # TODO do this check?
                        raise ValueError("out_dtype has too many indices")
                    return_ = self._out_dtype[0][1](return_)

            return return_

        # /def
        return wrapper

    # /def


# /class

# -------------------------------------------------------------------

# define class
class dtypeDecoratorBase:
    """Ensure arguments are type `dtype`.

    Parameters
    ----------
    func : function (optional)
        function to decorate
    inargs : Ellipsis or iterable or slice or None (optional)
        - None (default), does nothing
        - Ellipsis: converts all arguments to dtype
        - iterable: convert arguments at index specified in iterable
            ex: [0, 2] converts arguments 0 & 2
        - slice: convert all arguments specified by slicer
    outargs : Ellipsis or iterable or slice or None (optional)
        - None (default), does nothing
        - iterable: convert arguments at index specified in iterable
            ex: [0, 2] converts arguments 0 & 2
        - slice: convert all arguments specified by slicer
        - Ellipsis : convert all arguments

    these arguments, except func, should be specified by key word.
    if inargs is forgotten and func is not a function, then func is
    assumed to be inargs.

    Examples
    --------
    >>> intDecorator = dtypeDecoratorMaker(int)
    >>> @intDecorator(inargs=[0, 1], outargs=2)
    ... def func(x, y, z):
    ...     return x, y, z, (x, y, z)
    ... # /def
    >>> print(func(1.1, 2.2, 3.3))
    (1, 2, 3, (1, 2, 3.3))

    """

    def __init_subclass__(cls, dtype: T.Any = None):
        """Initialize subclass & store dtype."""
        super().__init_subclass__()
        cls._dtype = dtype

    # /def

    @property
    def dtype(self):
        """Get dtype. read-only access."""
        return self._dtype

    # /def

    def __new__(
        cls,
        func: T.Callable = None,
        inargs: T.Union[EllipsisType, slice, T.Iterable, None] = None,
        outargs: T.Union[EllipsisType, slice, T.Iterable, None] = None,
    ):
        self = super().__new__(cls)  # making instance of self

        # correcting if forgot to specify inargs and did not provide a func
        # in this case, *inargs* is stored in *func*
        # need to do func->None, inarags<-func, and outargs<-inargs
        if not isinstance(func, T.Callable):
            # moving arguments 'forward'
            outargs = inargs
            inargs = func
            func = None

        # allowing for wrapping with calling the class
        if func is not None:
            # need to initialize b/c not returning `self`
            self.__init__(inargs, outargs)
            return self(func)
        else:
            return self

    # /def

    def __init__(
        self,
        inargs: T.Union[EllipsisType, slice, T.Iterable, None] = None,
        outargs: T.Union[EllipsisType, slice, T.Iterable, None] = None,
    ) -> None:
        super().__init__()

        # inargs
        if inargs is Ellipsis:  # convert all
            self._inargs = slice(None)
        else:
            self._inargs = inargs

        # TODO validate inputs

        # outargs
        if outargs is Ellipsis:
            self._outargs = slice(None)
        elif np.isscalar(outargs):
            self._outargs = [outargs]
        else:
            self._outargs = outargs

    # /def

    def __call__(self, wrapped_func: T.Callable) -> T.Callable:
        """Wrap function.

        Works by making a wrapper which will convert input and
        output arguments to the specified data type.

        Parameters
        ----------
        wrapped_func : callable
            Function to wrap.

        """
        sig = inspect.signature(wrapped_func)

        @functools.wraps(wrapped_func)
        def wrapper(*args: T.Any, **kwargs: T.Any) -> T.Any:

            ba = sig.bind_partial(*args, **kwargs)
            ba.apply_defaults()

            # PRE
            # making arguments self._dtype
            if self._inargs is None:  # no conversion needed
                pass
            elif isinstance(self._inargs, slice):
                # converting inargs to list of indices
                lna = len(ba.args)

                inkeys = tuple(ba.arguments.keys())[:lna][self._inargs]
                inargs = tuple(range(lna))[self._inargs]

                # converting to desired dtype
                for k, i in zip(inkeys, inargs):
                    ba.arguments[k] = self._dtype(ba.args[i])
            else:  # any iterable

                lna = len(ba.args)
                argkeys = tuple(ba.arguments.keys())

                for i in self._inargs:
                    if isinstance(i, int):  # it's for args
                        ba.arguments[argkeys[i]] = self._dtype(args[i])
                    else:  # isinstance(i, str)
                        ba.arguments[i] = self._dtype(ba.arguments[i])

            # /PRE

            return_ = wrapped_func(*ba.args, **ba.kwargs)

            # POST
            # no conversion needed
            if self._outargs is None or not isinstance(return_, tuple):
                return return_
            # slice
            elif isinstance(self._outargs, slice):
                iterable = tuple(range(len(return_)))[self._outargs]
            else:
                iterable = self._outargs

            return_ = list(return_)

            for i in iterable:
                return_[i] = self._dtype(return_[i])

            return tuple(return_)
            # /POST

        # /def

        return wrapper

    # /def


# /class


# -------------------------------------------------------------------


def dtypeDecoratorMaker(dtype: T.Any):
    """Function to make a dtype decorator.

    Parameters
    ----------
    dtype : type
        intended data type

    Returns
    -------
    dtypeDecorator : decorator class
        a decorator which can convert input and output arguments
        to the intended datatype

    Examples
    --------
    >>> intDecorator = dtypeDecoratorMaker(int)
    >>> @intDecorator(inargs=[0, 1], outargs=2)
    ... def func(x, y, z):
    ...     return x, y, z, (x, y, z)
    >>> x, y, z, orig = func(1.1, 2.2, 3.3)
    >>> print(x, y, z, orig)  # z->int before returned
    1 2 3 (1, 2, 3.3)

    """
    # make subclass
    class dtypeDecorator(dtypeDecoratorBase, dtype=dtype):
        pass

    # /class

    dtypeDecorator.__name__ = f"{dtype}Decorator"

    return dtypeDecorator


# /def


#############################################################################
# MAKING DECORATORS

intDecorator = dtypeDecoratorMaker(int)
floatDecorator = dtypeDecoratorMaker(float)
strDecorator = dtypeDecoratorMaker(str)
boolDecorator = dtypeDecoratorMaker(bool)

ndarrayDecorator = dtypeDecoratorMaker(np.ndarray)
ndfloat64Decorator = dtypeDecoratorMaker(np.float64)


#############################################################################
# END



# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""Setup for :mod:`~utilipy.imports`."""


###############################################################################
# IMPORTS

from __future__ import absolute_import

# BUILT-IN
import typing as T
from contextlib import contextmanager

# THIRD PARTY
from astropy import config as _config

###############################################################################
# CODE
###############################################################################


class Conf(_config.ConfigNamespace):
    """Configuration parameters for `utilipy.imports`."""

    verbose_imports = _config.ConfigItem(
        True,
        description="imports import verbosity configuration",
        cfgtype="boolean(default=True)",
    )


conf = Conf()
# /class


# -------------------------------------------------------------------


@contextmanager
def use_import_verbosity(verbosity: T.Optional[bool] = None):
    """Set import verbosity only inside a with block.

    Wrapper for conf.set_temp('verbose_imports', verbosity).

    Parameters
    ----------
    verbosity : bool, Optional
        the import verbosity, None (default) doesn't change value.

    Yields
    ------
    :class:`~utilipy.imports.conf`

    """
    with conf.set_temp("verbose_imports", verbosity):  # use existing system.
        yield conf

    return


# /def


###############################################################################
# INFO


__all__ = [
    "conf",  # configuration instance
    "use_import_verbosity",  # contextmanager
]


###############################################################################
# END



# -*- coding: utf-8 -*-

# ----------------------------------------------------------------------------
#
# TITLE   : numpy Parse tools
# PROJECT : utilipy
#
# ----------------------------------------------------------------------------

# Docstring
"""Numpy parse tools.

routines to parse docstrings according to the Napoleon format specifications.
code sourced from [1]_, but modified to merge, not replace, docstring sections.


Routine Listings
----------------
merge_numpy_docs
    Merge two numpy-style docstrings into a single docstring

References
----------
The code source [1]_.

.. [1] https://github.com/rsokl/custom_inherit

TODO
----
https://stackoverflow.com/a/41083968


"""

__author__ = "Ryan Soklaski"
__license__ = "MIT"
__version__ = "2.2.2"
__email__ = "https://github.com/rsokl/custom_inherit"

__all__ = ["merge_numpy_docs"]


###############################################################################
# IMPORTS

# BUILT-IN
import typing as T
from collections import OrderedDict
from inspect import cleandoc

# THIRD PARTY
from typing_extensions import Literal

###############################################################################
# CODE
###############################################################################


def parse_numpy_doc(doc):
    """Extract the text from the sections of a numpy-formatted docstring.

    Parameters
    ----------
    doc: Union[str, None]

    Returns
    -------
    OrderedDict[str, Union[None,str]]
        The extracted numpy-styled docstring sections.

    """
    doc_sections = OrderedDict(
        [
            ("Short Summary", None),
            ("Deprecation Warning", None),
            ("Attributes", None),
            ("Extended Summary", None),
            ("Parameters", None),
            ("Returns", None),
            ("Yields", None),
            ("Other Parameters", None),
            ("Raises", None),
            ("See Also", None),
            ("Notes", None),
            ("References", None),
            ("Examples", None),
        ]
    )

    if not doc:
        return doc_sections

    doc = cleandoc(doc)
    lines = iter(doc.splitlines())

    key = "Short Summary"
    body = []
    while True:
        try:
            line = next(lines).rstrip()
            if line in doc_sections:
                doc_sections[key] = "\n".join(body).rstrip() if body else None
                body = []
                key = line
                next(lines)  # skip section delimiter
            else:
                body.append(line)
        except StopIteration:
            doc_sections[key] = "\n".join(body)
            break

    return doc_sections


# /def


def merge_section(
    key: str,
    prnt_sec: T.Union[str, None],
    child_sec: T.Union[str, None],
    method: T.Union[Literal["merge"], Literal["replace"]] = "replace",
) -> T.Union[str, None]:
    """Synthesize a output numpy docstring section.

    Parameters
    ----------
    key: str
        The numpy-section being merged.
    prnt_sec: Optional[str]
        The docstring section from the parent's attribute.
    child_sec: Optional[str]
        The docstring section from the child's attribute.
    method: {merge, replace}, optional
        'merge' merges the sections.
        'replace' (default) parent with child sections, if exists.

    Returns
    -------
    Optional[str]
        The output docstring section.

    """
    if prnt_sec is None and child_sec is None:
        return None

    if key == "Short Summary":
        header = ""
    else:
        header = "\n".join((key, "".join("-" for i in range(len(key))), ""))

    if method == "replace":
        body = prnt_sec if child_sec is None else child_sec
    elif method == "merge":
        # TODO more intelligent merge strategy,
        # respecting arg vs kwarg order, overriding repeated args, etc.
        if prnt_sec is None:
            body = child_sec or ""
        elif child_sec is None:
            body = prnt_sec or ""
        else:
            body = prnt_sec + "\n" + child_sec

    return header + body


# /def


def merge_all_sections(
    prnt_sctns: T.MutableMapping[str, T.Union[None, str]],
    child_sctns: T.MutableMapping[str, T.Union[None, str]],
    method: T.Union[Literal["merge"], Literal["replace"]] = "replace",
) -> T.Union[str, None]:
    """Merge the doc-sections of the parent's and child's attribute.

    Parameters
    ----------
    prnt_sctns: OrderedDict[str, Union[None,str]]
    child_sctns: OrderedDict[str, Union[None,str]]
    method: {merge, replace}, optional
        'merge' merges the sections.
        'replace' (default) parent with child sections, if exists.

    Returns
    -------
    str
        Output docstring of the merged docstrings.

    """
    doc = []

    prnt_only_raises = prnt_sctns["Raises"] and not (
        prnt_sctns["Returns"] or prnt_sctns["Yields"]
    )
    if prnt_only_raises and (child_sctns["Returns"] or child_sctns["Yields"]):
        prnt_sctns["Raises"] = None

    for key in prnt_sctns:
        sect = merge_section(
            key, prnt_sctns[key], child_sctns[key], method=method
        )
        if sect is not None:
            doc.append(sect)

    return "\n\n".join(doc) if doc else None


# /def


def merge_numpy_docs(
    prnt_doc: T.Union[str, None] = None,
    child_doc: T.Union[str, None] = None,
    method: T.Union[Literal["merge"], Literal["replace"]] = "replace",
) -> T.Union[str, None]:
    """Merge two numpy-style docstrings into a single docstring.

    Given the numpy-style docstrings from a parent and child's attributes,
    merge the docstring sections such that the child's section is used,
    wherever present, otherwise the parent's section is used.

    Any whitespace that can be uniformly removed from a docstring's second line
    and onwards is removed. Sections will be separated by a single blank line.

    Parameters
    ----------
    prnt_doc: Optional[str]
        The docstring from the parent.
    child_doc: Optional[str]
        The docstring from the child.
    method: {merge, replace}, optional
        'merge' merges the sections.
        'replace' (default) parent with child sections, if exists.

    Returns
    -------
    Union[str, None]
        The merged docstring.

    """
    return merge_all_sections(
        parse_numpy_doc(prnt_doc), parse_numpy_doc(child_doc), method=method
    )


# /def


###############################################################################
# END



# -*- coding: utf-8 -*-

"""Functions for making basic pickling easier."""

__author__ = "Nathaniel Starkman"


__all__ = [
    "dump",
    "dump_many",
    "load",
    "load_many",
]

#############################################################################
# IMPORTS

# BUILT-IN
import pickle as _pickle
import typing as T
import warnings

# THIRD PARTY
from astropy import config as _config
from astropy.utils.decorators import format_doc

# PROJECT-SPECIFIC
from .logging import LogFile

try:
    import dill
except ImportError:
    HAS_DILL = False
else:
    HAS_DILL = True


#############################################################################
# CONFIGURATION


class Conf(_config.ConfigNamespace):
    """Configuration parameters for :mod:`~utilipy.utils.exceptions`."""

    use_dill = _config.ConfigItem(
        False,
        description="When True, try to use `dill` instead of `pickle`.",
        cfgtype="boolean(default=False)",
    )


conf = Conf()
# /class

# Warn if want to but cannot use dill
if conf.use_dill and not HAS_DILL:

    warnings.warn("`dill` cannot be imported. Will use `pickle` instead.")


#############################################################################
# PARAMETERS

_LOGFILE = LogFile(header=False)


#############################################################################
# CODE


@format_doc(None, odoc="\n\t".join(_pickle.dump.__doc__.split("\n")))
def dump(
    obj: T.Any,
    fname: str,
    protocol: T.Optional[int] = None,
    *,
    fopt: str = "b",
    fix_imports: bool = True,
    use_dill: T.Optional[bool] = None,
    # logger
    logger: LogFile = _LOGFILE,
    verbose: T.Optional[int] = None,
) -> None:
    """Wrap pickle.dump.

    `fname` replaces *file* and is a string for the filename
    this file is auto opened and closed

    pickle.dump docstring:

        {odoc}

    """
    if use_dill is None:
        use_dill = conf.use_dill

    if use_dill and not HAS_DILL:
        raise ValueError("dill is not installed. cannot use dill.")

    logger.report(
        f"dumping obj at {fname}",
        (
            f"dumping obj at {fname} with fopt='{'w' + fopt}, "
            f"protocol=protocol, fix_imports={fix_imports}'"
        ),
        verbose=verbose,
    )

    with open(fname, "w" + fopt) as file:
        if use_dill:
            dill.dump(obj, file, protocol=protocol, fix_imports=fix_imports)
        else:
            _pickle.dump(obj, file, protocol=protocol, fix_imports=fix_imports)

    return


# /def


save = dump


# --------------------------------------------------------------------------


def dump_many(
    *objs: T.Tuple[T.Any, str],
    protocol: T.Any = None,
    fopt: str = "b",
    fix_imports: bool = True,
    use_dill=None,
    # logger
    logger: LogFile = _LOGFILE,
    verbose: T.Optional[int] = None,
) -> None:
    """Wrap pickle.dump.

    Parameters
    ----------
    *objs: Tuple[Any, str]
    protocol: Any
        (default None)
    fopt: str
        (default "b")
    fix_imports: bool
        (default True)
    logger: LogFile
    verbose: int or None, optional

    """
    for (obj, fname) in objs:
        dump(
            obj=obj,
            fname=fname,
            protocol=protocol,
            fopt=fopt,
            fix_imports=fix_imports,
            use_dill=use_dill,
            # logger
            logger=logger,
            verbose=verbose,
        )

    return


# /def


##########################################################################


@format_doc(None, odoc="\n\t".join(_pickle.load.__doc__.split("\n")))
def load(
    *fnames: str,
    fopt: str = "b",
    fix_imports: bool = True,
    encoding: str = "ASCII",
    errors: str = "strict",
    use_dill=None,
    # logger
    logger: LogFile = _LOGFILE,
    verbose: T.Optional[int] = None,
):
    """Wrap pickle.load.

    *fname* replaces *file* and is a string for the filename
    this file is auto opened and closed

    pickle.load docstring:

        {odoc}

    """
    if use_dill is None:
        use_dill = conf.use_dill

    if use_dill and not HAS_DILL:
        raise ValueError("dill is not installed. cannot use dill.")

    res: list = [None] * len(fnames)  # preload results list
    for i, fname in enumerate(fnames):  # iterate through files

        logger.report(
            f"loading obj at {fname}",
            (
                f"loading obj at {fname} with fopt='{'r' + fopt}, "
                f"fix_imports={fix_imports}', encoding={encoding}, "
                f"errors={errors}"
            ),
            verbose=verbose,
        )

        with open(fname, "r" + fopt) as file:
            if use_dill:
                res[i] = dill.load(
                    file,
                    fix_imports=fix_imports,
                    encoding=encoding,
                    errors=errors,
                )
            else:
                res[i] = _pickle.load(
                    file,
                    fix_imports=fix_imports,
                    encoding=encoding,
                    errors=errors,
                )

    if len(fnames) == 1:
        return res[0]
    return res


# /def

load_many = load


# --------------------------------------------------------------------------

# def save_pickles(savefilename, *args, **kwargs):
#     """
#     NAME:
#        save_pickles
#     PURPOSE:
#        relatively safely save things to a pickle
#     INPUT:
#        savefilename - name of the file to save to; actual save operation
#                       will be performed on a temporary file and then that
#                       file will be shell mv'ed to savefilename
#        +things to pickle (as many as you want!)
#     OUTPUT:
#        none
#     HISTORY:
#        2010-? - Written - Bovy (NYU)
#        2011-08-23 - generalized and added to galpy.util - Bovy (NYU)

#     TODO work into this python

#     """
#     saving = True
#     interrupted = False
#     file, tmp_savefilename = tempfile.mkstemp()  # savefilename+'.tmp'
#     os.close(file)  # Easier this way
#     while saving:
#         try:
#             savefile = open(tmp_savefilename, 'wb')
#             file_open = True
#             if kwargs.get('testKeyboardInterrupt', False) and not interrupted:
#                 raise KeyboardInterrupt
#             for f in args:
#                 pickle.dump(f, savefile, pickle.HIGHEST_PROTOCOL)
#             savefile.close()
#             file_open = False
#             shutil.move(tmp_savefilename, savefilename)
#             saving = False
#             if interrupted:
#                 raise KeyboardInterrupt
#         except KeyboardInterrupt:
#             if not saving:
#                 raise
#             print("KeyboardInterrupt ignored while saving pickle ...")
#             interrupted = True
#         finally:
#             if file_open:
#                 savefile.close()
# # /def


#############################################################################
# END



# -*- coding: utf-8 -*-

"""Tests for :mod:`~utilipy.utils.exceptions`."""


__all__ = [
    "test_utilipyWarning",
    "test_utilipyWarningVerbose",
]


##############################################################################
# IMPORTS

# BUILT-IN
import sys
import tempfile
import unittest
import warnings

# PROJECT-SPECIFIC
from utilipy.utils import exceptions

##############################################################################
# Tests


def test_utilipyWarning():
    """Test :class:`~utilipy.utils.exceptions.utilipyWarning`."""
    self = unittest.TestCase()
    with self.assertWarns(exceptions.utilipyWarning):
        warnings.warn("utilipyWarning", exceptions.utilipyWarning)

    return


# /def


def test_utilipyWarningVerbose():
    """Test :class:`~utilipy.utils.exceptions.utilipyWarningVerbose`."""
    self = unittest.TestCase()
    with self.assertWarns(exceptions.utilipyWarningVerbose):
        warnings.warn(
            "utilipyWarningVerbose", exceptions.utilipyWarningVerbose
        )

    return


# /def


# --------------------------------------------------------------------------


def test_showwarning():
    """Test :func:`~utilipy.utils.exceptions._showwarning`."""
    # First test UserWarning
    with tempfile.NamedTemporaryFile(mode="w+") as tmp:

        exceptions.showwarning(
            message="\ntest",
            category=UserWarning,
            filename=tmp.name,
            lineno=-1,
            file=tmp,  # redirect to file, not stdout
            line=None,
        )

        tmp.seek(0)  # rewind file

        lines = tmp.readlines()
        i = 1 if sys.version_info.minor > 6 else 3

        assert lines[i] == "test\n", lines

    # /with

    # Test utilipyWarning non-verbose warning
    # have to make a temporary file
    # turn off the verbosity, and pass a utilipyWarning
    with tempfile.NamedTemporaryFile(mode="w+") as tmp:

        with exceptions.conf.set_temp("verbose_warnings", False):

            exceptions.showwarning(
                message="test",
                category=exceptions.utilipyWarning,
                filename=tmp.name,
                lineno=-1,
                file=tmp,  # redirect to file, not stdout
                line=None,
            )

            tmp.seek(0)  # rewind file

            assert tmp.readline() == "utilipyWarning: test\n", tmp.readline()

        # /with

    # /with


# /def


##############################################################################
# END



# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE.rst


"""Decorators for code in active development.

See Also
--------
deprecation decorators in :function:`~astropy.util.exceptions.deprecated`,
upon which this code is based.

"""

__all__ = [
    "indev",
    "indev_doc",
    "indev_attribute",
]


##############################################################################
# IMPORTS

# BUILT-IN
import functools
import inspect
import textwrap
import types
import typing as T
import warnings

# PROJECT-SPECIFIC
from utilipy.utils.exceptions import utilipyWarning

##############################################################################
# PARAMETERS

method_types = (classmethod, staticmethod, types.MethodType)

##############################################################################
# CODE
##############################################################################


class DevelopmentWarning(utilipyWarning):
    """Warning class to indicate a feature as being actively developed.

    Expect the feature to change in input, output and implementation.

    """


# /class


class BetaDevelopmentWarning(DevelopmentWarning):
    """Warning class to indicate an upcoming feature that is nearly ready."""


# /class

##############################################################################


def _get_function(func):
    """Get function object given function.

    Returns
    -------
    func : Callable
        The function from the function wrapper.

    """
    if isinstance(func, method_types):
        func = func.__func__
    return func


# /def


# -------------------------------------------------------------------


def indev_function(func: T.Callable, message: str, warning_type):
    """Wrap function to display ``warning_type`` when called.

    Parameters
    ----------
    func : Callable
    message : str
        The warning message
    warning_type : type
        Warning class.

    Returns
    -------
    Callable
        Wrapped function.

    """
    if isinstance(func, method_types):
        func_wrapper = type(func)
    else:

        def func_wrapper(func):
            return func

    func = _get_function(func)

    def upcoming_func(*args, **kwargs):
        warnings.warn(message, warning_type, stacklevel=2)

        return func(*args, **kwargs)

    # If this is an extension function, we can't call
    # functools.wraps on it, but we normally don't care.
    # This crazy way to get the type of a wrapper descriptor is
    # straight out of the Python 3.3 inspect module docs.
    if type(func) is not type(str.__dict__["__add__"]):  # noqa
        upcoming_func = functools.wraps(func)(upcoming_func)

    upcoming_func.__doc__ = indev_doc(upcoming_func.__doc__, message)

    return func_wrapper(upcoming_func)


# /def


# -------------------------------------------------------------------


def indev_class(cls, message: str, warning_type):
    """Class in-dev.

    Update the docstring and wrap the ``__init__`` in-place (or ``__new__``
    if the class or any of the bases overrides ``__new__``) so it will give
    an in-dev warning when an instance is created.

    This won't work for extension classes because these can't be modified
    in-place and the alternatives don't work in the general case:

    - Using a new class that looks and behaves like the original doesn't
      work because the __new__ method of extension types usually makes sure
      that it's the same class or a subclass.
    - Subclassing the class and return the subclass can lead to problems
      with pickle and will look weird in the Sphinx docs.

    Returns
    -------
    cls : ClassType

    """
    cls.__doc__ = indev_doc(cls.__doc__, message)

    if cls.__new__ is object.__new__:
        cls.__init__ = indev_function(
            _get_function(cls.__init__),
            message=message,
            warning_type=warning_type,
        )
    else:
        cls.__new__ = indev_function(
            _get_function(cls.__new__),
            message=message,
            warning_type=warning_type,
        )

    return cls


# /def

##############################################################################


def indev_doc(old_doc: T.Union[str, None], message: str):
    """Returns a given docstring with an in-dev message prepended.

    .. todo::

        process docstring better

    Parameters
    ----------
    old_doc: str
    message : str

    Returns
    -------
    new_doc : str

    """
    if not old_doc:  # empty str or None
        old_doc = ""

    old_doc = textwrap.dedent(old_doc).strip("\n")

    note_indev = (
        "\n    .. versionchanged:: indev\n"
        "\n        {message}\n".format(**{"message": message.strip()})
    )

    # TODO, process docstring better
    _sections = [
        "\n    Parameters\n    ----------",
        "\n    Returns\n    -------",
    ]
    if _sections[0] in old_doc:  # check if has Parameters
        descr, rest = old_doc.split(_sections[0])
        descr += note_indev
        new_doc = descr + _sections[0] + rest

    elif _sections[1] in old_doc:  # maybe starts with Returns
        descr, rest = old_doc.split(_sections[1])
        descr += note_indev
        new_doc = descr + _sections[1] + rest

    else:  # just a regular doc
        new_doc = old_doc + "\n" + note_indev + "\n"

    return new_doc


# /def


# -------------------------------------------------------------------


def indev(
    message: T.Union[str, T.Callable] = "",
    name: str = "",
    alternative: str = "",
    todo: str = "",
    obj_type: T.Optional[str] = None,
    warning_type=DevelopmentWarning,
):
    """Used to mark a function or class as in the development phase.

    Expect the feature to change in input, output and implementation.

    To mark an attribute as upcoming / in development, use `indev_attribute`.

    Parameters
    ----------
    message : str, optional
        Override the default in-dev message.  The format
        specifier ``func`` may be used for the name of the function,
        and ``alternative`` may be used in the in-dev message
        to insert the name of an alternative to the in-dev
        function. ``obj_type`` may be used to insert a friendly name
        for the type of object being in-dev.

    name : str, optional
        The name of the in-dev function or class; if not provided
        the name is automatically determined from the passed in
        function or class, though this is useful in the case of
        renamed functions, where the new function is just assigned to
        the name of the in-dev function.  For example::

            def new_function():
                ...
            oldFunction = new_function

    alternative : str, optional
        An alternative function or class name that the user may use in
        place of the in-dev object.  The in-dev warning will
        tell the user about this alternative if provided.

    todo : str, optional
        A todo about intended features and functionality. The in-dev warning
        will tell the user about this alternative if provided.

    obj_type : str, optional
        The type of this object, if the automatically determined one
        needs to be overridden.

    warning_type : warning
        Warning to be issued.
        Default is `~DevelopmentWarning`.


    See Also
    --------
    deprecation decorators in :func:`~astropy.util.exceptions.deprecated`

    """
    # wrapper function
    def make_indev(
        obj,
        message=message,
        name=name,
        alternative=alternative,
        warning_type=warning_type,
    ):
        """Mark object as in-development.

        Parameters
        ----------
        obj : object
        message : str
        name : str
        alternative : str
        warning_type : ClassType

        Returns
        -------
        Callable

        """
        if obj_type is None:
            if isinstance(obj, type):
                obj_type_name = "class"
            elif inspect.isfunction(obj):
                obj_type_name = "function"
            elif inspect.ismethod(obj) or isinstance(obj, method_types):
                obj_type_name = "method"
            else:
                obj_type_name = "object"
        else:
            obj_type_name = obj_type

        if not name:
            name = _get_function(obj).__name__

        altmessage = ""
        todomessage = ""
        if not message or (type(message) is type(make_indev)):
            message = (
                "The {func} {obj_type} is in development and may "
                "be added in a future version."
            )
            if alternative:
                altmessage = f"\n        Use {alternative} instead."
            if todo:
                todomessage = f"\n        TODO: {todo}"

        message = (
            message.format(
                **{
                    "func": name,
                    "name": name,
                    "alternative": alternative,
                    "todo": todo,
                    "obj_type": obj_type_name,
                }
            )
            + altmessage
            + todomessage
        )

        if isinstance(obj, type):  # decorate class
            return indev_class(obj, message=message, warning_type=warning_type)
        else:  # decorate function
            return indev_function(
                obj, message=message, warning_type=warning_type
            )

    # /def

    if type(message) is type(make_indev):
        return make_indev(message)

    return make_indev


# /def


# ------------------------------------------------------------------------


def indev_attribute(
    name,
    message=None,
    alternative=None,
    todo=None,
    warning_type=DevelopmentWarning,
):
    """Mark a public attribute as in development.

    This creates a property that will warn when the given attribute name
    is accessed. To prevent the warning (i.e. for internal code),
    use the private name for the attribute by prepending an underscore
    (i.e. ``self._name``).

    Parameters
    ----------
    name : str
        The name of the in-dev attribute.

    message : str, optional
        Override the default in-dev message.  The format
        specifier ``name`` may be used for the name of the attribute,
        and ``alternative`` may be used in the in-dev message
        to insert the name of an alternative to the in-dev
        function.

    alternative : str, optional
        An alternative attribute that the user may use in place of the
        in-dev attribute.  The in-dev warning will tell the
        user about this alternative if provided.

    todo : str, optional
        A todo about intended functionality. The in-dev warning
        will tell the user about this alternative if provided.

    warning_type : warning
        Warning to be issued.
        Default is `~DevelopmentWarning`.

    Examples
    --------
    ::

        class MyClass:
            # Mark the new_attr as in development
            new_attr = misc.indev_attribute('new_attr')

            def method(self):
                self._new_attr = 42

    Returns
    -------
    property

    """
    private_name = "_" + name

    @indev(
        name=name,
        obj_type="attribute",
        warning_type=warning_type,
        alternative=alternative,  # TODO, implement for attribute
        todo=todo,  # TODO, implement for attribute
    )
    def get(self):
        return getattr(self, private_name)

    @indev(
        name=name,
        obj_type="attribute",
        warning_type=warning_type,
        alternative=alternative,  # TODO, implement for attribute
        todo=todo,  # TODO, implement for attribute
    )
    def set(self, val):
        setattr(self, private_name, val)

    @indev(
        name=name,
        obj_type="attribute",
        warning_type=warning_type,
        alternative=alternative,  # TODO, implement for attribute
        todo=todo,  # TODO, implement for attribute
    )
    def delete(self):
        delattr(self, private_name)

    return property(get, set, delete)


# /def


##############################################################################
# END



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for :mod:`~utilipy.decorators.func_io.dtype_decorator`."""


##############################################################################
# IMPORTS

# THIRD PARTY
import numpy as np
import pytest

# PROJECT-SPECIFIC
from utilipy.decorators.func_io import dtypeDecorator

##############################################################################
# PARAMETERS

x = 1
y = 2.2
z = [3, 4.4]
u, uu = "5", "5.5"
v = False
w = np.array([7, 8.8])


##############################################################################
# dtypeDecorator


def test_dtypeDecorator_blank():
    """Test no-conversion mode of dtypeDecorator."""
    # defining function
    @dtypeDecorator()
    def func(x):
        return x

    # /def

    assert func(x) == x
    assert func(y) == y
    assert func(z) == z
    assert func(u) == u
    assert func(uu) == uu
    assert func(v) == v
    assert (func(w) == w).all()

    return


# /def


def test_dtypeDecorator_python_scalars():
    """Test standard use of dtypeDecorator."""
    # defining function
    @dtypeDecorator(in_dtype=[(0, int)], out_dtype=[(0, float)])
    def func(x):
        assert isinstance(x, int)
        return x

    # /def

    assert isinstance(func(x), float)
    assert isinstance(func(y), float)

    # should fail
    with pytest.raises(TypeError):
        assert isinstance(func(z), float)

    assert isinstance(func(u), float)

    with pytest.raises(ValueError):
        assert isinstance(func(uu), float)

    assert isinstance(func(v), float)

    with pytest.raises(TypeError):
        assert isinstance(func(w), float)

    return


# /def


# def test_dtypeDecorator_numpy_arrays():
#     """test standard use of dtypeDecorator
#     """
#     # TODO
# # /def


# def test_dtypeDecorator_single_arg():
#     """
#     """
#
#     @dtypeDecorator(inargs=(0, int))
#     def func(x):
#         return x
#
#     # TODO test
#
#     @dtypeDecorator(outargs=(0, int))
#     def func(x):
#         return x
#
#     # TODO test
#
#     @dtypeDecorator(inargs=(0, int), outargs=(1, int))
#     def func(x, y):
#         return x, y
#
#     # TODO test
#
#     return
# # /def


# def test_dtypeDecorator_string_arg():
#     """
#     """
#
#     @dtypeDecorator(inargs=('all', int))
#     def func(x):
#         return x
#
#     # TODO test
#
#     @dtypeDecorator(outargs=('all', int))
#     def func(x):
#         return x
#
#     # TODO test
#
#     @dtypeDecorator(inargs=('all', int), outargs=('all', float))
#     def func(x, y):
#         return x, y
#
#     # TODO test
#
#     return
# # /def


# --------------------------------------------------------------------------


##############################################################################
# END



# -*- coding: utf-8 -*-

# ----------------------------------------------------------------------------
#
# TITLE   : AMUSE imports
# PROJECT : utilipy
#
# ----------------------------------------------------------------------------

"""Base set of imports for :mod:`~amuse`.

Routine Listings
----------------
Amuse: imports

    - amuse
    - amuse.lab
    - amuse.units.units, constants
    - amuse.couple.bridge

Notes
-----
If you are using this import, please cite all the references from the codes
you used (AMUSE will show you which papers to cite for these),
and the version of AMUSE you were using
(https://doi.org/10.5281/zenodo.3260650)

.. image:: https://zenodo.org/badge/DOI/10.5281/zenodo.3260650.svg
   :target: https://doi.org/10.5281/zenodo.3260650


References
----------
AMUSE citations [#]_, [#]_, [#]_, [#]_

.. [#] Portegies Zwart, S. and McMillan, S.L.W., 2018, “Astrophysical Recipes:
   the art of AMUSE", AAS IOP Astronomy publishing (411 pages).

.. [#] Portegies Zwart, S. et al., 2013, "Multi-physics Simulations Using a
   Hierarchical Interchangeable Software Interface", Computer Physics
   Communications 183, 456-468 (2013CoPhC.183..456P).

.. [#] Pelupessy, F. I. et al., 2013, "The Astrophysical Multipurpose Software
   Environment", Astronomy and Astrophysics 557, 84 (2013AandA…557A..84P).

.. [#] Portegies Zwart, S. et al., 2009, "A multiphysics and multiscale
   software environment for modeling astrophysical systems", New Astronomy,
   Volume 14, Issue 4, 369-378 (2009NewA…14..369P).

"""

__author__ = "Nathaniel Starkman"


__all__ = [
    "amuse_imports_help",
]


##############################################################################
# HELPER FUNCTIONS

# PROJECT-SPECIFIC
from utilipy.imports import conf
from utilipy.utils import make_help_function

##############################################################################
# IMPORTS

try:

    import amuse

except ImportError:

    import warnings

    warnings.warn("Cannot import amuse")

else:

    from amuse import lab
    from amuse.units import units, constants
    from amuse.couple import bridge

    __all__ += [
        "amuse",
        "lab",
        "units",
        "constants",
        "bridge",
    ]

# /if

##############################################################################
# Printing Information

amuse_imports_help = make_help_function(
    "amuse", __doc__, look_for="Routine Listings"
)


if conf.verbose_imports:
    amuse_imports_help()


##############################################################################
# END



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""test functions for galpy imports."""


__all__ = ["test_import_amuse"]

##############################################################################
# IMPORTS

# BUILT-IN
import warnings

# THIRD PARTY
import pytest

try:
    import galpy
except ImportError:
    HAS_GALPY = False
else:
    if isinstance(galpy.__version__, str):  # check not empty namespace
        HAS_GALPY = True
    else:
        HAS_GALPY = False


##############################################################################
# CODE
##############################################################################


@pytest.mark.skipif(~HAS_GALPY, reason="requires galpy")
def test_import_amuse():
    """Test AMUSE imports."""
    warnings.filterwarnings("ignore", category=DeprecationWarning)
    from utilipy.imports import amuse_imports as imports

    for obj in (
        "galpy",
        "potential",
        "MWPotential2014",
        "Orbit",
        "bovy_conversion",
        "bovy_coords",
    ):
        assert hasattr(imports, obj)


# /def


##############################################################################
# END



# -*- coding: utf-8 -*-

"""Tests for :mod:`~utilipy.math.core`."""


__all__ = [
    "test_quadrature_no_input",
    "test_quadrature_single_argument",
    "test_quadrature_multi_scalar_argument",
    "test_quadrature_single_vector_argument",
    "test_quadrature_multi_vector_argument",
    "test_as_quantity_unchanged",
    "test_as_quantity_modifications",
    "test_as_quantity_recast",
    "test_as_quantity_exceptions",
    "test_qsquare",
    "test_qnorm",
    "test_qarange",
]


##############################################################################
# IMPORTS

# BUILT-IN
import os

# THIRD PARTY
import astropy.units as u
import numpy as np
import pytest

# PROJECT-SPECIFIC
from utilipy.math import core

##############################################################################
# PARAMETERS

_NP_V = [int(v) for i, v in enumerate(np.__version__.split(".")) if i < 3]


##############################################################################
# CODE
##############################################################################

# -------------------------------------------------------------------
# Quadrature


def test_quadrature_no_input():
    """Test :class:`~utilipy.math.core.quadrature` no input Exception."""
    with pytest.raises(ValueError):
        core.quadrature()  # No input


# /def


def test_quadrature_single_argument():
    """Test :class:`~utilipy.math.core.quadrature` single input."""
    assert core.quadrature(2.0) == 2.0
    assert core.quadrature(-2.0) == 2.0

    # axis argument can matter
    with pytest.raises(np.AxisError):
        assert core.quadrature(-2.0, axis=1) == 2.0

    with pytest.raises(np.AxisError):
        core.quadrature(2, axis=2)


# /def


def test_quadrature_multi_scalar_argument():
    """Test :class:`~utilipy.math.core.quadrature` multi-scalar inputs."""
    assert core.quadrature(3.0, 4.0) == 5.0
    assert core.quadrature(-3, 4.0) == 5.0

    # axis argument can matter
    assert core.quadrature(3.0, 4.0, axis=-1) == 5.0

    with pytest.raises(np.AxisError):
        core.quadrature(3.0, 4.0, axis=2)


# /def


def test_quadrature_single_vector_argument():
    """Test :class:`~utilipy.math.core.quadrature` single vector inputs."""
    x = [3.0, 12.0]

    # not expanded
    actual = core.quadrature(x)
    expected = 12.36931687685298
    np.testing.assert_almost_equal(actual, expected, decimal=7)

    # same as expanded
    actual = core.quadrature(*x)
    np.testing.assert_equal(actual, expected)

    # axis argument can matter
    actual = core.quadrature(x, axis=-1)
    np.testing.assert_equal(actual, expected)

    with pytest.raises(np.AxisError):
        core.quadrature(x, axis=2)


# /def


def test_quadrature_multi_vector_argument():
    """Test :class:`~utilipy.math.core.quadrature` multi-vector argument."""
    x = [3.0, 12.0]
    y = [4.0, 5.0]

    actual = core.quadrature(x, y)
    expected = np.array([5.0, 13.0])
    np.testing.assert_equal(actual, expected)

    # axis argument matters
    actual = core.quadrature(x, y, axis=-1)
    expected = np.array([12.36931688, 6.40312424])
    np.testing.assert_almost_equal(actual, expected)

    with pytest.raises(np.AxisError):
        core.quadrature(x, y, axis=2)


# /def


# -------------------------------------------------------------------
# As_quantity


def test_as_quantity_unchanged():
    """Test :func:`~utilipy.math.core.as_quantity` pass-through."""
    # single number
    x = 1 * u.m
    y = core.as_quantity(x)
    assert y == x

    # array
    x = [1, 2] * u.m
    y = core.as_quantity(x)
    assert all(y == x)


# /def


def test_as_quantity_modifications():
    """Test :func:`~utilipy.math.core.as_quantity` in-place modification."""
    # array
    x = [1, 2, 3] * u.m
    y = core.as_quantity(x)
    assert all(y == x)

    # changed
    x[0] = 0 * u.m
    assert all(y == x)


# /def


def test_as_quantity_recast():
    """Test :func:`~utilipy.math.core.as_quantity` array recasting."""
    # array
    x = [1 * u.m, 2 * u.m, 3 * u.m]
    y = core.as_quantity(x)
    assert all(y == [1, 2, 3] * u.m)

    # array, with conversion
    x = [1 * u.m, 200 * u.cm, 3 * u.m]
    y = core.as_quantity(x)
    assert all(y == [1, 2, 3] * u.m)


# /def


def test_as_quantity_exceptions():
    """Test :func:`~utilipy.math.core.as_quantity` expected Exceptions."""
    with pytest.raises(NotImplementedError):
        core.as_quantity("arg")


# /def


# -------------------------------------------------------------------


def test_qsquare():
    """Test :func:`~utilipy.math.core.qsquare`.

    Most tests are covered by `test_as_quantity`
    and numpy's internal testing for :func:`~numpy.square`

    """
    x = 2 * u.m
    y = core.qsquare(x)
    assert y == x ** 2


# /def


# -------------------------------------------------------------------


def test_qnorm():
    """Test :func:`~utilipy.math.core.qnorm`.

    Most tests are covered by `test_as_quantity`
    and numpy's internal testing for :func:`~numpy.linalg.norm`

    """
    # Have to split into pre and post numpy 1.16
    # https://docs.astropy.org/en/lts/whatsnew/4.0.html
    if (
        (os.environ.get("NUMPY_EXPERIMENTAL_ARRAY_FUNCTION", 0) == 0)
        & (_NP_V[0] <= 1)
        & (_NP_V[1] <= 16)
    ):
        # norm a scalar
        x = -2 * u.m
        y = core.qnorm(x)
        assert y == 2.0

        # norm an array
        x = [3, 4] * u.m
        y = core.qnorm(x)
        assert y == 5.0

    else:

        # norm a scalar
        x = -2 * u.m
        y = core.qnorm(x)
        assert y == -x

        # norm an array
        x = [3, 4] * u.m
        y = core.qnorm(x)
        assert y == 5 * u.m


# /def


# -------------------------------------------------------------------


def test_qarange():
    """Test :func:`~utilipy.math.core.qarange`.

    Most tests are covered by `test_as_quantity`
    and numpy's internal testing for :func:`~numpy.arange`

    """
    # basic test
    start = 1 * u.m
    stop = 10 * u.m
    step = 1 * u.m

    expected = np.arange(1, 10, 1) * u.m
    assert all(core.qarange(start, stop, step) == expected)

    # change unit
    expected = np.arange(1, 10, 1) * 100 * u.cm
    assert all(core.qarange(start, stop, step, unit=u.cm) == expected)

    # change step
    step = 1 * u.cm
    expected = np.arange(100, 1000, 1) * u.cm
    assert all(core.qarange(start, stop, step) == expected)

    # raise error
    with pytest.raises(AttributeError):
        core.qarange(start, stop, 1)


# /def


##############################################################################
# END



# -*- coding: utf-8 -*-

# ----------------------------------------------------------------------------
#
# TITLE   : ReST Parse tools
# PROJECT : utilipy
#
# ----------------------------------------------------------------------------

# Docstring
"""ReST parse tools.

routines to parse docstrings according to the Napoleon format specifications.
code sourced from [1]_, but modified to merge, not replace, docstring sections.


Routine Listings
----------------
merge_rest_docs
    Merge two ReST-style docstrings into a single docstring

References
----------
The code source [1]_.

.. [1] https://github.com/rsokl/custom_inherit

"""

__all__ = ["merge_rest_docs"]


###############################################################################
# IMPORTS

# BUILT-IN
import typing as T
from collections import OrderedDict
from inspect import cleandoc
from string import punctuation

# THIRD PARTY
from typing_extensions import Literal

###############################################################################
# CODE
###############################################################################


def is_delimiter(line: str) -> bool:
    """True if a line consists only of a single punctuation character."""
    return (
        bool(line)
        and (line[0] in punctuation)
        and (line[0] * len(line) == line)
    )


# /def


class Section(object):
    def __init__(self, header: str = None, body: str = None):
        self.header: str = header
        self.body: str = body


# /class


def parse_rest_doc(doc: T.Union[str, None]) -> T.Dict[str, Section]:
    """Extract the headers, delimiters, and text from reST-formatted docstrings.

    Parameters
    ----------
    doc: Union[str, None]

    Returns
    -------
    Dict[str, Section]

    """
    doc_sections = OrderedDict([("", Section(header=""))])
    if not doc:
        return doc_sections

    doc = cleandoc(doc)
    lines = iter(doc.splitlines())

    header = ""
    body = []
    section = Section(header=header)
    line = ""

    while True:
        try:
            prev_line = line
            line = next(lines)
            # section header encountered
            if is_delimiter(line) and 0 < len(prev_line) <= len(line):
                # prev-prev-line is overline
                if (
                    len(body) >= 2
                    and len(body[-2]) == len(line)
                    and body[-2][0] == line[0]
                    and is_delimiter(body[-2])
                ):
                    lim = -2
                else:
                    lim = -1

                section.body = "\n".join(body[:lim]).rstrip()
                doc_sections.update([(header.strip(), section)])
                section = Section(header="\n".join(body[lim:] + [line]))
                header = prev_line
                body = []
                line = ""
            else:
                body.append(line)

        except StopIteration:
            section.body = "\n".join(body).rstrip()
            doc_sections.update([(header.strip(), section)])
            break

    return doc_sections


# /def


def merge_rest_docs(
    prnt_doc: T.Optional[str] = None,
    child_doc: T.Optional[str] = None,
    method: T.Union[Literal["merge"], Literal["replace"]] = "replace",
) -> str:
    """See custom_inherit.style_store.reST for details."""
    prnt_sections = parse_rest_doc(prnt_doc)
    child_sections = parse_rest_doc(child_doc)

    header = prnt_sections[""]
    prnt_sections.update(child_sections)
    if not child_sections[""].body:
        prnt_sections[""] = header
        if not header.body:
            prnt_sections.popitem(last=False)

    return "\n\n".join(
        ("\n".join((x.header, x.body)) for x in prnt_sections.values())
    ).lstrip()


# /def


###############################################################################
# END
